import os
from typing import Any, List, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
import scanpy as sc
import sklearn.metrics
import torch
from anndata import AnnData, concat
from scdataloader import Collator, Preprocessor
from scdataloader.data import SimpleAnnDataset
from scdataloader.utils import get_descendants, random_str
from scipy.optimize import curve_fit
from scipy.stats import spearmanr
from simpler_flash import FlashTransformer
from torch.nn import functional as F
from torch.utils.data import DataLoader
from tqdm import tqdm

from scprint2.model import utils

FILE_DIR = os.path.dirname(os.path.realpath(__file__))


class Denoiser:
    def __init__(
        self,
        batch_size: int = 10,
        num_workers: int = 1,
        max_len: int = 5_000,
        how: str = "most var",
        max_cells: int = 500_000,
        doplot: bool = False,
        predict_depth_mult: int = 4,
        downsample_expr: Optional[float] = None,
        genelist: Optional[List[str]] = None,
        save_every: int = 100_000,
        pred_embedding: List[str] = ["cell_type_ontology_term_id"],
        additional_info: bool = False,
        apply_zero_pred: bool = False,
        use_knn: bool = True,
    ):
        """
        Denoiser class for denoising scRNA-seq data using a scPRINT model

        Args:
            batch_size (int, optional): Batch size for processing. Defaults to 10.
            num_workers (int, optional): Number of workers for data loading. Defaults to 1.
            max_len (int, optional): Maximum number of genes to consider. Defaults to 5000.
            how (str, optional): Method to select genes. Options are "most var", "random expr", "some". Defaults to "most var".
                - "most var": select the most variable genes
                - "random expr": select random expressed genes
                - "some": select a subset of genes defined in genelist
            max_cells (int, optional): Number of cells to use for plotting correlation. Defaults to 10000.
            doplot (bool, optional): Whether to generate plots of the similarity between the denoised and true expression data. Defaults to False.
                Only works when downsample_expr is not None and max_cells < 100.
            predict_depth_mult (int, optional): Multiplier for prediction depth. Defaults to 4.
                This will artificially increase the sequencing depth (or number of counts) to 4 times the original depth.
            downsample_expr (Optional[float], optional): Fraction of expression data to downsample. Defaults to None.
                This is usefull to test the ability of the model to denoise the dataset. only to use the input data as a benchmark dataset.
                When this option is on, the denoiser will output benchmark metrics
            genelist (List[str], optional): The list of genes to be used for embedding. Defaults to []: In this case, "how" needs to be "most var" or "random expr".
            save_every (int, optional): The number of cells to save at a time. Defaults to 100_000.
                This is important to avoid memory issues.
            pred_embedding (List[str], optional): The embedding type to be used as the denoising will also predict the cell embeddings.
            additional_info (bool, optional): Whether to print additional benchmark information during denoising. Defaults to False.
                only useful when downsampling is used.
            apply_zero_pred (bool, optional): Whether to apply zero inflation to the output value during denoising, else uses only the predicted mean.
                applying zero inflation might give results closer to the specific biases of sequencing technologies but less biological truthful.
            use_knn (bool, optional): Whether to use knn cells for denoising when the model uses metacell expression embedding. Defaults to True.
        """
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.max_len = max_len
        self.max_cells = max_cells
        self.doplot = doplot
        self.predict_depth_mult = predict_depth_mult
        self.how = how
        self.downsample_expr = downsample_expr
        self.genelist = genelist
        self.save_every = save_every
        self.pred_embedding = pred_embedding
        self.additional_info = additional_info
        self.apply_zero_pred = apply_zero_pred
        self.use_knn = use_knn

    def __call__(self, model: torch.nn.Module, adata: AnnData) -> tuple[dict, Optional[np.ndarray], AnnData]:
        """
        __call__ calling the function

        Args:
            model (torch.nn.Module): The scPRINT model to be used for denoising.
            adata (AnnData): The annotated data matrix of shape n_obs x n_vars. Rows correspond to cells and columns to genes.

        Returns:
            dict: The benchmark metrics if downsampling is used.
            Optional[np.ndarray]: The random set of cells used if max_cells < adata.shape[0].
            AnnData: The denoised annotated data matrix.
        """
        # Select random number
        random_indices = None
        if self.max_cells < adata.shape[0]:
            random_indices = np.random.randint(
                low=0, high=adata.shape[0], size=self.max_cells
            )
            adataset = SimpleAnnDataset(
                adata[random_indices],
                obs_to_output=["organism_ontology_term_id"],
                get_knn_cells=model.expr_emb_style == "metacell" and self.use_knn,
            )
        else:
            adataset = SimpleAnnDataset(
                adata,
                obs_to_output=["organism_ontology_term_id"],
                get_knn_cells=model.expr_emb_style == "metacell" and self.use_knn,
            )
        if self.how == "most var":
            sc.pp.highly_variable_genes(
                adata, flavor="seurat_v3", n_top_genes=self.max_len, span=0.99
            )
            self.genelist = adata.var.index[adata.var.highly_variable]
        else:
            self.genelist = adata.var.index
        self.genelist = [i for i in model.genes if i in self.genelist]
        print(f"working on {len(self.genelist)} accepted genes")

        col = Collator(
            organisms=model.organisms,
            valid_genes=model.genes,
            max_len=self.max_len,
            how="some" if self.how == "most var" else self.how,
            genelist=self.genelist if self.how != "random expr" else [],
            n_bins=model.n_input_bins if model.expr_emb_style == "binned" else 0,
        )
        dataloader = DataLoader(
            adataset,
            collate_fn=col,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=False,
        )

        prevplot = model.doplot
        model.doplot = self.doplot
        model.on_predict_epoch_start()
        model.eval()
        device = model.device.type
        model.pred_log_adata = True
        stored_noisy = None
        rand = random_str()
        dtype = (
            torch.float16
            if type(model.transformer) is FlashTransformer
            else model.dtype
        )
        torch.cuda.empty_cache()
        save_expr = model.save_expr
        model.save_expr = True
        with torch.no_grad(), torch.autocast(device_type=device, dtype=dtype):
            for batch in tqdm(dataloader):
                gene_pos, expression, depth = (
                    batch["genes"].to(device),
                    batch["x"].to(device),
                    batch["depth"].to(device),
                )
                knn_cells = (
                    batch["knn_cells"].to(device)
                    if model.expr_emb_style == "metacell" and self.use_knn
                    else None
                )
                if self.downsample_expr is not None:
                    expression = utils.downsample_profile(
                        expression, self.downsample_expr
                    )
                    if knn_cells is not None:
                        for i in range(knn_cells.shape[1]):
                            knn_cells[:, i] = utils.downsample_profile(
                                knn_cells[:, i], self.downsample_expr
                            )
                if stored_noisy is None:
                    stored_noisy = expression.cpu().numpy()
                else:
                    stored_noisy = np.concatenate(
                        [stored_noisy, expression.cpu().numpy()], axis=0
                    )

                model._predict(
                    gene_pos,
                    expression,
                    depth,
                    knn_cells=(
                        batch["knn_cells"].to(device)
                        if model.expr_emb_style == "metacell" and self.use_knn
                        else None
                    ),
                    knn_cells_info=(
                        batch["knn_cells_info"].to(device)
                        if model.expr_emb_style == "metacell" and self.use_knn
                        else None
                    ),
                    do_generate=False,
                    depth_mult=self.predict_depth_mult,
                    pred_embedding=self.pred_embedding,
                    max_size_in_mem=self.save_every,
                    name="denoise_" + rand + "_",
                )
        torch.cuda.empty_cache()
        model.log_adata(name="denoise_" + rand + "_" + str(model.counter))
        try:
            mdir = (
                model.logger.save_dir if model.logger.save_dir is not None else "data"
            )
        except:
            mdir = "data"
        pred_adata = []
        for i in range(model.counter + 1):
            file = (
                mdir
                + "/step_"
                + str(model.global_step)
                + "_"
                + model.name
                + "_denoise_"
                + rand
                + "_"
                + str(i)
                + "_"
                + str(model.global_rank)
                + ".h5ad"
            )
            pred_adata.append(sc.read_h5ad(file))
            os.remove(file)
        pred_adata = concat(pred_adata)

        if model.transformer.attn_type == "hyper":
            # seq len must be a multiple of 128
            num = (1 if model.use_metacell_token else 0) + (
                (len(model.classes) + 1) if not model.cell_transformer else 0
            )
            if (stored_noisy.shape[1] + num) % 128 != 0:
                stored_noisy = stored_noisy[
                    :, : ((stored_noisy.shape[1]) // 128 * 128) - num
                ]
        pred_adata.X = stored_noisy

        metrics = None
        model.doplot = prevplot
        model.save_expr = save_expr
        if self.downsample_expr is not None:
            reco = np.array(pred_adata.layers["scprint_mu"].data).reshape(
                pred_adata.shape[0], -1
            )
            # reco = reco * F.sigmoid(
            #    torch.Tensor(np.array(pred_adata.layers["scprint_pi"].data).reshape(pred_adata.shape[0], -1)) < 0.5
            # ).numpy()

            adata = (
                adata[random_indices, adata.var.index.isin(pred_adata.var.index)]
                if random_indices is not None
                else adata[:, adata.var.index.isin(pred_adata.var.index)]
            )
            true = adata[
                :,
                pred_adata.var.index[
                    pred_adata.var.index.isin(adata.var.index)
                ].to_list(),
            ].X.toarray()
            if self.apply_zero_pred:
                reco = (
                    reco
                    * (
                        1
                        - F.sigmoid(
                            torch.Tensor(
                                np.array(pred_adata.layers["scprint_pi"].data).reshape(
                                    pred_adata.shape[0], -1
                                )
                            )
                        )
                    ).numpy()
                )

            corr_coef, p_value = spearmanr(
                np.vstack([reco[true != 0], stored_noisy[true != 0], true[true != 0]]).T
            )
            metrics = {
                "reco2noisy": corr_coef[0, 1],
                "reco2full": corr_coef[0, 2],
                "noisy2full": corr_coef[1, 2],
            }
            if self.additional_info:
                # Sample only 3000 elements for correlation calculation
                if reco.shape[0] > 3000:
                    indices = np.random.choice(reco.shape[0], 3000, replace=False)
                    reco = reco[indices]
                    stored_noisy = stored_noisy[indices]
                    true = true[indices]
                corr, p_value = spearmanr(
                    np.vstack(
                        [
                            reco.flatten(),
                            stored_noisy.flatten(),
                            true.flatten(),
                        ]
                    ).T
                )
                m = {
                    "reco2full": corr[0, 2],
                    "noisy2full": corr[1, 2],
                }
                print("corr with zeros: ")
                print(m)
                cell_wise = np.array(
                    [
                        spearmanr(reco[i][true[i] != 0], true[i][true[i] != 0])[0]
                        for i in range(reco.shape[0])
                    ]
                )
                torm = np.array(
                    [
                        spearmanr(stored_noisy[i][true[i] != 0], true[i][true[i] != 0])[
                            0
                        ]
                        for i in range(reco.shape[0])
                    ]
                )
                cell_wise -= torm
                cell_wise_zero = np.mean(
                    [spearmanr(reco[i], true[i])[0] for i in range(reco.shape[0])]
                )
                print("cell_wise self corr (reco, noisy, true)")
                print(
                    {
                        "cell_wise_w_zero": cell_wise_zero,
                        "cell_wise_to_noisy": np.mean(cell_wise),
                    }
                )
                print("depth-wise plot")
                plot_cell_depth_wise_corr_improvement(cell_wise, (true > 0).sum(1))

            if self.doplot and self.max_cells < 100:
                corr_coef[p_value > 0.05] = 0
                plt.figure(figsize=(10, 5))
                plt.imshow(
                    corr_coef, cmap="coolwarm", interpolation="none", vmin=-1, vmax=1
                )
                plt.colorbar()
                plt.title("Expression Correlation Coefficient")
                plt.show()
        return metrics, random_indices, pred_adata


# testdatasets=['/R4ZHoQegxXdSFNFY5LGe.h5ad', '/SHV11AEetZOms4Wh7Ehb.h5ad',
# '/V6DPJx8rP3wWRQ43LMHb.h5ad', '/Gz5G2ETTEuuRDgwm7brA.h5ad', '/YyBdEsN89p2aF4xJY1CW.h5ad',
# '/SO5yBTUDBgkAmz0QbG8K.h5ad', '/r4iCehg3Tw5IbCLiCIbl.h5ad', '/SqvXr3i3PGXM8toXzUf9.h5ad',
# '/REIyQZE6OMZm1S3W2Dxi.h5ad', '/rYZ7gs0E0cqPOLONC8ia.h5ad', '/FcwMDDbAQPNYIjcYNxoc.h5ad',
# '/fvU5BAMJrm7vrgDmZM0z.h5ad', '/gNNpgpo6gATjuxTE7CCp.h5ad'],


def default_benchmark(
    model: Any,
    folder_dir: str = FILE_DIR + "/../../data/",
    dataset: str = FILE_DIR
    + "/../../data/gNNpgpo6gATjuxTE7CCp.h5ad",  # r4iCehg3Tw5IbCLiCIbl
) -> dict:
    """
    default_benchmark function used to run the default denoising benchmark of scPRINT

    Args:
        model (Any): The scPRINT model to be used for the benchmark.
        folder_dir (str, optional): Directory containing data files.
        dataset (str, optional): Path to the dataset to use for benchmarking.

    Returns:
        dict: A dictionary containing the benchmark metrics.
    """
    if dataset.startswith("https://"):
        adata = sc.read(
            folder_dir + dataset.split("/")[-1],
            backup_url=dataset,
        )
    else:
        adata = sc.read_h5ad(dataset)
    if dataset.split("/")[-1] == "gNNpgpo6gATjuxTE7CCp.h5ad":
        use_layer = "counts"
        is_symbol = True
    else:
        use_layer = None
        is_symbol = False
    max_len = 4000 if adata.X.sum(1).mean() < 150_000 else 8000
    preprocessor = Preprocessor(
        use_layer=use_layer,
        is_symbol=is_symbol,
        force_preprocess=True,
        skip_validate=True,
        do_postp=model.expr_emb_style == "metacell",
        drop_non_primary=False,
    )
    adata = preprocessor(adata.copy())
    if model.expr_emb_style == "metacell":
        if "X_pca" not in adata.obsm:
            sc.pp.pca(adata, n_comps=50)
        sc.pp.neighbors(adata, use_rep="X_pca")
    denoise = Denoiser(
        batch_size=40 if model.expr_emb_style != "metacell" else 20,
        max_len=max_len,
        max_cells=10_000,
        doplot=False,
        num_workers=8,
        predict_depth_mult=5,
        downsample_expr=0.7,
        pred_embedding=model.pred_embedding,
    )
    return denoise(model, adata)[0]


def mse(test_data, denoised_data, target_sum=1e4):
    sc.pp.normalize_total(test_data, target_sum=target_sum)
    sc.pp.log1p(test_data)

    sc.pp.normalize_total(denoised_data, target_sum=target_sum)
    sc.pp.log1p(denoised_data)

    print("Compute mse value", flush=True)
    return sklearn.metrics.mean_squared_error(test_data.X.todense(), denoised_data.X)


# from molecular_cross_validation.mcv_sweep import poisson_nll_loss
# copied from: https://github.com/czbiohub/molecular-cross-validation/blob/master/src/molecular_cross_validation/mcv_sweep.py
def poisson_nll_loss(y_pred: np.ndarray, y_true: np.ndarray) -> float:
    return (y_pred - y_true * np.log(y_pred + 1e-6)).mean()


def split_molecules(
    umis: np.ndarray,
    data_split: float,
    overlap_factor: float = 0.0,
    random_state: np.random.RandomState = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """Splits molecules into two (potentially overlapping) groups.
    :param umis: Array of molecules to split
    :param data_split: Proportion of molecules to assign to the first group
    :param overlap_factor: Overlap correction factor, if desired
    :param random_state: For reproducible sampling
    :return: umis_X and umis_Y, representing ``split`` and ``~(1 - split)`` counts
             sampled from the input array
    """
    if random_state is None:
        random_state = np.random.RandomState()

    umis_X_disjoint = random_state.binomial(umis, data_split - overlap_factor)
    umis_Y_disjoint = random_state.binomial(
        umis - umis_X_disjoint, (1 - data_split) / (1 - data_split + overlap_factor)
    )
    overlap_factor = umis - umis_X_disjoint - umis_Y_disjoint
    umis_X = umis_X_disjoint + overlap_factor
    umis_Y = umis_Y_disjoint + overlap_factor

    return umis_X, umis_Y


def plot_cell_depth_wise_corr_improvement(corr_coef, y):
    def linear_func(x, a, b):
        return a * np.log(x) + b

    # Fit the linear curve
    ppot, _ = curve_fit(linear_func, y, corr_coef)

    # Plot the data points
    plt.scatter(
        y, corr_coef, label="denoising increase as depth increase", color="blue"
    )

    # Plot the fitted linear curve
    x_values = np.linspace(min(y), max(y), 100)
    plt.plot(
        x_values,
        linear_func(x_values, *ppot),
        label="Linear Fit",
        color="red",
        linestyle="--",
    )

    plt.xlabel("True sum (depth)")
    plt.ylabel("Denoising improvement")
    plt.ylim(-0.5, 1)
    plt.legend()
    plt.show()



import math
from typing import Optional

import numpy as np
import torch
from torch import Tensor, nn
from torch.nn import functional as F
from torch_geometric.data import Batch, Data
from torch_geometric.nn import MLP, GATConv, GCNConv, SAGEConv
from torch_geometric.nn.aggr import DeepSetsAggregation


class GeneEncoder(nn.Module):
    def __init__(
        self,
        num_embeddings: int,
        embedding_dim: int,
        padding_idx: Optional[int] = None,
        weights: Optional[Tensor] = None,
        weights_file: Optional[str] = None,
        freeze: bool = False,
    ):
        """
        Encodes gene sequences into a continuous vector space using an embedding layer.
        Uses memory mapping for efficient access to large embedding files.

        Args:
            num_embeddings (int): The number of possible values
            embedding_dim (int): The dimension of the output vectors
            padding_idx (int, optional): The index of the padding token
            weights (Tensor, optional): The initial weights for the embedding layer
            weights_file (str, optional): Path to parquet file containing embeddings
            freeze (bool, optional): Whether to freeze the weights of the embedding layer
        """
        super(GeneEncoder, self).__init__()
        self.output_dim = embedding_dim

        if weights_file is not None:
            self.memmap = True
            if not freeze:
                raise ValueError(
                    "freeze must be True when using memory-mapped embeddings"
                )
            # Load the parquet file and create memory-mapped array
            import os

            import pandas as pd

            # Create memory-mapped file path
            self.mmap_file = f"{weights_file}.mmap"
            self.loc = None
            self.enc = None
            # Only create the memory-mapped file if it doesn't exist
            if not os.path.exists(self.mmap_file):
                print(f"Creating memory-mapped file for embeddings at {self.mmap_file}")
                df = pd.read_parquet(weights_file)
                embeddings = torch.nn.AdaptiveAvgPool1d(self.output_dim)(
                    torch.tensor(df.values)
                )

                # Create memory-mapped array
                self.embeddings = np.memmap(
                    self.mmap_file, dtype="float32", mode="w+", shape=embeddings.shape
                )
                # Copy data to memory-mapped array
                self.embeddings[:] = embeddings.numpy()
                #
                self.embeddings.flush()

                # Clean up memory
                del df
                del embeddings
            else:
                print(
                    f"Loading existing memory-mapped embeddings from {self.mmap_file}"
                )
                # Load existing memory-mapped file
                self.embeddings = np.memmap(
                    self.mmap_file,
                    dtype="float32",
                    mode="r",  # Read-only mode since we don't need to modify
                    shape=(num_embeddings, embedding_dim),
                )
        else:
            self.memmap = False
            self.embeddings = nn.Embedding(
                num_embeddings, embedding_dim, padding_idx=padding_idx, _freeze=freeze
            )
            if weights is not None:
                self.embeddings.weight.data.copy_(torch.Tensor(weights))

    def forward(self, x: Tensor) -> Tensor:
        """
        Forward pass of the encoder.

        Args:
            x (Tensor): Input tensor of indices [batch_size, seq_len]

        Returns:
            Tensor: Embedded vectors [batch_size, seq_len, embedding_dim]
        """
        if self.memmap:
            if self.loc is None or not torch.all(x.sum(1) == self.loc):
                self.enc = (
                    torch.from_numpy(
                        self.embeddings[x.reshape(-1).cpu().numpy()].copy()
                    )
                    .reshape(x.shape + (-1,))
                    .to(x.device)
                )
                self.loc = x.sum(1)
            return self.enc.clone()
        else:
            return self.embeddings(x)

    def __del__(self):
        """Cleanup method to ensure proper handling of memory-mapped file."""
        if hasattr(self, "embeddings") and self.embeddings is not None:
            try:
                self.embeddings._mmap.close()
            except:
                pass

    def _init_weights(self):
        pass


class PositionalEncoding(nn.Module):
    def __init__(
        self,
        d_model: int,
        gene_pos_enc: list[str] = [],
    ):
        """
        The PositionalEncoding module applies a positional encoding to a sequence of vectors.
        This is necessary for the Transformer model, which does not have any inherent notion of
        position in a sequence. The positional encoding is added to the input embeddings and
        allows the model to attend to positions in the sequence.

        Args:
            d_model (int): The dimension of the input vectors.
            gene_pos_enc (list[str], optional): The gene position encoding to use.

        Note: not used in the current version of scprint-2.
        """
        super(PositionalEncoding, self).__init__()
        self.gene_pos_enc = gene_pos_enc
        max_len = max(gene_pos_enc)
        position = torch.arange(max_len).unsqueeze(1)
        token_to_pos = {token: pos for token, pos in enumerate(gene_pos_enc)}

        # Create a dictionary to convert token to position

        div_term = torch.exp(
            torch.arange(0, d_model, 2) * (-math.log(float(10_000)) / d_model)
        )
        pe = torch.zeros(max_len, 1, d_model)
        pe[:, 0, 0::2] = torch.sin(position * div_term)
        pe[:, 0, 1::2] = torch.cos(position * div_term)
        # we reorder them and map them to gene_id (position)
        arr = []
        for _, v in token_to_pos.items():
            arr.append(pe[v - 1].numpy())
        pe = torch.Tensor(np.array(arr))
        # Remove the unnecessary middle dimension since pe should be [m, d]
        # pe = pe.squeeze(1)
        self.register_buffer("pe", pe)

    def forward(self, gene_pos: Tensor) -> Tensor:
        """
        Args:
            gene_pos (Tensor): Gene position indices, shape [seq_len, batch_size] or [seq_len]

        Returns:
            Tensor: Positional encodings, shape [*gene_pos.shape, embedding_dim]
        """
        return torch.index_select(self.pe, 0, gene_pos.reshape(-1)).reshape(
            gene_pos.shape + (-1,)
        )


class DPositionalEncoding(nn.Module):
    """
    The PositionalEncoding module applies a positional encoding to a sequence of vectors.
    This is necessary for the Transformer model, which does not have any inherent notion of
    position in a sequence. The positional encoding is added to the input embeddings and
    allows the model to attend to positions in the sequence.

    Args:
        d_model (int): The dimension of the input vectors.
        max_len_x (int): The maximum length in the x dimension.
        max_len_y (int): The maximum length in the y dimension.
        maxvalue_x (float, optional): Maximum value for x dimension scaling. Defaults to 10000.0.
        maxvalue_y (float, optional): Maximum value for y dimension scaling. Defaults to 10000.0.

    Note: not used in the current version of scprint-2.
    """

    def __init__(
        self,
        d_model: int,
        max_len_x: int,
        max_len_y: int,
        maxvalue_x=10000.0,
        maxvalue_y=10000.0,
    ):
        super(DPositionalEncoding, self).__init__()
        position2 = torch.arange(max_len_y).unsqueeze(1)
        position1 = torch.arange(max_len_x).unsqueeze(1)

        half_n = d_model // 2

        div_term2 = torch.exp(
            torch.arange(0, half_n, 2) * (-math.log(maxvalue_y) / d_model)
        )
        div_term1 = torch.exp(
            torch.arange(0, half_n, 2) * (-math.log(maxvalue_x) / d_model)
        )
        pe1 = torch.zeros(max_len_x, 1, d_model)
        pe2 = torch.zeros(max_len_y, 1, d_model)
        pe1[:, 0, 0:half_n:2] = torch.sin(position1 * div_term1)
        pe1[:, 0, 1:half_n:2] = torch.cos(position1 * div_term1)
        pe2[:, 0, half_n::2] = torch.sin(position2 * div_term2)
        pe2[:, 0, 1 + half_n :: 2] = torch.cos(position2 * div_term2)
        # https://github.com/tatp22/multidim-positional-encoding/blob/master/positional_encodings/torch_encodings.py
        self.register_buffer("pe1", pe1)
        self.register_buffer("pe2", pe2)

        # PE(x,y,2i) = sin(x/10000^(4i/D))
        # PE(x,y,2i+1) = cos(x/10000^(4i/D))
        # PE(x,y,2j+D/2) = sin(y/10000^(4j/D))
        # PE(x,y,2j+1+D/2) = cos(y/10000^(4j/D))

    def forward(self, x: Tensor, pos_x: Tensor, pos_y: Tensor) -> Tensor:
        """
        Args:
            x: Tensor, shape [seq_len, batch_size, embedding_dim]
        """
        x = x + self.pe1[pos_x]
        x = x + self.pe2[pos_y]
        return x


class ContinuousValueEncoder(nn.Module):
    def __init__(
        self,
        d_model: int,
        dropout: float = 0.1,
        max_value: int = 100_000,
        layers: int = 1,
        size: int = 1,
    ):
        """
        Encode real number values to a vector using neural nets projection.

        Args:
            d_model (int): The dimension of the input vectors.
            dropout (float, optional): The dropout rate to apply to the output of the positional encoding.
            max_value (int, optional): The maximum value of the input. Defaults to 100_000.
            layers (int, optional): The number of layers in the encoder. Defaults to 1.
            size (int, optional): The size of the input. Defaults to 1.
        """
        super(ContinuousValueEncoder, self).__init__()
        self.max_value = max_value
        self.encoder = nn.ModuleList()
        self.output_dim = d_model
        # self.mask_value = nn.Embedding(1, d_model)
        self.encoder.append(nn.Linear(size, d_model))
        for _ in range(layers - 1):
            self.encoder.append(nn.LayerNorm(d_model))
            self.encoder.append(nn.ReLU())
            self.encoder.append(nn.Dropout(p=dropout))
            self.encoder.append(nn.Linear(d_model, d_model))

    def forward(self, x: Tensor, mask: Tensor = None) -> Tensor:
        """
        Args:
            x: Tensor, shape [batch_size, seq_len]
        """
        # expand last dimension
        x = x.unsqueeze(-1)
        # use the mask embedding when x=-1
        # mask = (x == -1).float()
        x = torch.clamp(x, min=0, max=self.max_value)
        for val in self.encoder:
            x = val(x)
        if mask is not None:
            x = x.masked_fill_(mask.unsqueeze(-1), 0)
            # x = x.masked_fill_(mask.unsqueeze(-1), self.mask_value(0))
        return x

    def _init_weights(self):
        pass
        # for m in self.encoder:
        #    if isinstance(m, nn.Linear):
        #        torch.nn.init.eye_(m.weight)


class ExprBasedFT(nn.Module):
    def __init__(
        self,
        d_model: int,
        gene_encoder: nn.Module,
        expr_encoder: nn.Module = nn.Identity(),
        dropout: float = 0.1,
        layers: int = 2,
        intermediary_d: int = 256 + 64,
    ):
        """
        Encode real number values to a vector using neural nets projection.

        Args:
            d_model (int): The dimension of the input vectors.
            gene_encoder (nn.Module): The gene name encoder module.
            expr_encoder (nn.Module, optional): The expression encoder module. Defaults to nn.Identity.
            dropout (float, optional): The dropout rate to apply to the output of the positional encoding.
            layers (int, optional): The number of layers in the encoder. Defaults to 2.
            intermediary_d (int, optional): The dimension of the intermediary layers. Defaults to 256 + 64.

        """
        super(ExprBasedFT, self).__init__()
        self.encoder = nn.ModuleList()
        # self.mask_value = nn.Embedding(1, d_model)
        self.add_module("gene_encoder", gene_encoder)
        self.add_module("expr_encoder", expr_encoder)
        expr_shape, gene_shape = (
            self.expr_encoder.output_dim,
            self.gene_encoder.output_dim,
        )
        self.encoder.append(nn.Linear(expr_shape + gene_shape, intermediary_d))
        for i in range(layers - 1):
            self.encoder.append(nn.LayerNorm(intermediary_d))
            self.encoder.append(nn.ReLU())
            self.encoder.append(nn.Dropout(p=dropout))
            self.encoder.append(
                nn.Linear(intermediary_d, intermediary_d if i < layers - 2 else d_model)
            )

    def forward(
        self,
        gene_pos: Tensor,
        expr: Optional[Tensor] = None,
        mask: Optional[Tensor] = None,
        neighbors: Optional[Tensor] = None,
        neighbors_info: Optional[Tensor] = None,
    ) -> Tensor:
        """
        Args:
            gene_pos (Tensor[batch_size, seq_len]): Gene position indices
                input to the gene encoder
            expr (Tensor[batch_size, seq_len], Optional): Expression values
                input to the expression encoder
            mask (Tensor[batch_size, seq_len], Optional): Mask for the input
                input to the expression encoder
            neighbors (Tensor[batch_size, seq_len, n_neighbors], Optional): Neighbors indices
                input to the expression encoder when it is a GNN
            neighbors_info (Tensor[batch_size, seq_len, n_neighbors], Optional):
                optional additional information about the neighbors
                input to the expression encoder when it is a GNN
        """
        # expand last dimension
        if neighbors is None and expr is None:
            expr = torch.zeros(
                (gene_pos.shape[0], gene_pos.shape[1], self.expr_encoder.output_dim),
                dtype=torch.float32,
                device=gene_pos.device,
            )
            # if no expr information: consider that it is all masked
        else:
            expr = (
                self.expr_encoder(expr, mask=mask)
                if neighbors is None
                else self.expr_encoder(expr, neighbors, neighbors_info, mask=mask)
            )
        gene_pos = self.gene_encoder(gene_pos)
        x = torch.cat([expr, gene_pos], dim=-1)
        for val in self.encoder:
            x = val(x)
        return x

    def _init_weights(self):
        pass

    #    for m in self.encoder:
    #        if isinstance(m, nn.Linear):
    #            torch.nn.init.eye_(m.weight)
    #    self.expr_encoder._init_weights()


class CategoryValueEncoder(nn.Module):
    def __init__(
        self,
        num_embeddings: int,
        embedding_dim: int,
        padding_idx: Optional[int] = None,
    ):
        """
        Encodes categorical values into a vector using an embedding layer and layer normalization.

        Args:
            num_embeddings (int): The number of possible values.
            embedding_dim (int): The dimension of the output vectors.
            padding_idx (int, optional): The index of the padding token. Defaults to None.

        Note: not used in the current version of scprint-2.
        """
        super(CategoryValueEncoder, self).__init__()
        self.embedding = nn.Embedding(
            num_embeddings, embedding_dim, padding_idx=padding_idx
        )

    def forward(self, x: Tensor, mask: Tensor = None) -> Tensor:
        x = self.embedding(x.long())  # (batch, seq_len, embsize)
        if mask is not None:
            x = x.masked_fill(mask.unsqueeze(-1), 0)
        return x

    def _init_weights(self):
        pass


class EasyExprGNN(nn.Module):
    def __init__(
        self,
        self_dim: int = 64,
        output_dim: int = 32,
        self_layers: int = 2,
        dropout: float = 0.1,
        shared_layers: int = 2,
        neighbors_layers: int = 2,
    ):
        """
        Easy Expression Graph Neural Network

        The main GNN used in scPRINT-2 for expression encoding.
        It is inspired from the DeepSets architecture to aggregate neighbor information.

        Args:
            self_dim (int): Dimension of the self features
            output_dim (int): Output dimension
            self_layers (int): Number of layers for self features
            dropout (float): Dropout rate
            shared_layers (int): Number of shared layers
            neighbors_layers (int): Number of layers for neighbors features
        """
        super(EasyExprGNN, self).__init__()
        self.output_dim = output_dim
        self.self_dim = self_dim
        # neighbors
        self.neighbors_layers = nn.ModuleList()
        self.neighbors_layers.append(nn.Linear(2, self_dim // 2))
        for i in range(neighbors_layers - 1):
            self.neighbors_layers.append(nn.LayerNorm(self_dim // 2))
            self.neighbors_layers.append(nn.ReLU())
            self.neighbors_layers.append(nn.Dropout(p=dropout))
            self.neighbors_layers.append(nn.Linear(self_dim // 2, self_dim // 2))
        # self
        self.self_layers = nn.ModuleList()
        self.self_layers.append(nn.Linear(1, self_dim // 2))
        for i in range(self_layers - 1):
            self.self_layers.append(nn.LayerNorm(self_dim // 2))
            self.self_layers.append(nn.ReLU())
            self.self_layers.append(nn.Dropout(p=dropout))
            self.self_layers.append(nn.Linear(self_dim // 2, self_dim // 2))
        # shared
        self.shared_layers = nn.ModuleList()
        for i in range(shared_layers - 1):
            self.shared_layers.append(nn.Linear(self_dim, self_dim))
            self.shared_layers.append(nn.LayerNorm(self_dim))
            self.shared_layers.append(nn.ReLU())
            self.shared_layers.append(nn.Dropout(p=dropout))
        self.shared_layers.append(nn.Linear(self_dim, output_dim))

    def forward(
        self,
        expr: Optional[Tensor] = None,
        neighbors: Optional[Tensor] = None,
        edge_info: Optional[Tensor] = None,
        mask: Optional[Tensor] = None,
    ) -> Tensor:
        """
        Forward pass of the Easy Expression GNN

        Args:
            expr: Tensor of shape (batch, seq_len) representing expression values
            neighbors: Tensor of shape (batch, seq_len, n_neighbors) representing neighbor indices
            edge_info: Tensor of shape (batch, seq_len, n_neighbors) representing edge information
            mask: Tensor of shape (batch, seq_len) representing mask for the input

        Returns:
            Tensor of shape (batch, seq_len, output_dim) representing the output features
        """
        # batch, seq_len, neighbs
        if neighbors is None:
            neighbors = torch.zeros(
                (expr.shape[0], expr.shape[1], self.self_dim // 2), device=expr.device
            )
        else:
            neighbors = neighbors.transpose(1, 2)
            neighbors = torch.cat(
                [neighbors.unsqueeze(-1), edge_info.unsqueeze(-1)], dim=-1
            )
            for i, layer in enumerate(self.neighbors_layers):
                # batch, seq_len, neighbs, hidden_dim
                neighbors = layer(neighbors)
            neighbors = neighbors.sum(-2)
        if expr is None:
            expr = torch.zeros(
                (neighbors.shape[0], neighbors.shape[1], 1), device=neighbors.device
            )
        else:
            expr = expr.unsqueeze(-1)
            for i, layer in enumerate(self.self_layers):
                expr = layer(expr)
        x = torch.cat([expr, neighbors], dim=-1)
        for layer in self.shared_layers:
            # batch, seq_len, neighbs, hidden_dim
            x = layer(x)
        if mask is not None:
            x = x.masked_fill(mask.unsqueeze(-1), 0)
        return x

    def _init_weights(self):
        pass
        # for m in self.neighbors_layers:
        #    if isinstance(m, nn.Linear):
        #        torch.nn.init.zeros_(m.weight)
        #        if m.bias is not None:
        #            torch.nn.init.constant_(m.bias, 0)
        # for m in self.self_layers:
        #    if isinstance(m, nn.Linear):
        #        torch.nn.init.eye_(m.weight)
        #        if m.bias is not None:
        #            torch.nn.init.constant_(m.bias, 0)
        # for m in self.shared_layers:
        #    if isinstance(m, nn.Linear):
        #        torch.nn.init.eye_(m.weight)
        #        if m.bias is not None:
        #            torch.nn.init.constant_(m.bias, 0)
        #


class GNN(nn.Module):
    def __init__(
        self,
        input_dim: int = 1,  # here, 1 or 2
        merge_dim: int = 32,
        output_dim: int = 256,
        num_layers: int = 2,
        dropout: float = 0.1,
        gnn_type: str = "deepset",
        add_connection_feature: bool = False,
    ):
        """
        Graph Neural Network model

        Another implementation of a GNN layer that can be used for expression encoding.
        Supports GCN, GAT, GraphSAGE, and DeepSets architectures.

        Args:
            input_dim: Dimension of input node features
            output_dim: Dimension of output node features
            num_layers: Number of GNN layers
            dropout: Dropout probability
            gnn_type: Type of GNN layer ('gcn', 'gat', 'sage', or 'deepset')
        """
        super().__init__()

        self.input_dim = input_dim
        self.output_dim = output_dim
        if num_layers == 1:
            raise ValueError("num_layers must be greater than 1")
        self.num_layers = num_layers
        self.dropout = dropout
        self.gnn_type = gnn_type
        self.add_connection_feature = add_connection_feature

        if gnn_type == "deepset":
            # Local MLP (phi) for processing individual nodes
            self.input_nn_layer = MLP(
                in_channels=input_dim,
                hidden_channels=merge_dim,
                out_channels=merge_dim,
                num_layers=num_layers,
                dropout=dropout,
                act="relu",
                norm="layer_norm",
            )

            self.input_self_layer = MLP(
                in_channels=input_dim,
                hidden_channels=merge_dim + 2,
                out_channels=merge_dim,
                num_layers=num_layers - 1,
                dropout=dropout,
                act="relu",
                norm="layer_norm",
            )

            # Global MLP (rho) for processing aggregated features
            self.output_layer = MLP(
                in_channels=(
                    (merge_dim * 2) + 1 if add_connection_feature else merge_dim * 2
                ),
                hidden_channels=output_dim,
                out_channels=output_dim,
                num_layers=num_layers,
                dropout=dropout,
                act="relu",
                norm="layer_norm",
            )

            return

        # Select GNN layer type for other architectures
        else:
            if gnn_type == "gcn":
                gnn_layer = GCNConv
            elif gnn_type == "gat":
                gnn_layer = GATConv
            elif gnn_type == "sage":
                gnn_layer = SAGEConv
            else:
                raise ValueError(f"Unknown GNN type: {gnn_type}")

            self.gnn_layer = gnn_layer(
                output_dim,
                output_dim,
                add_self_loops=False,
                normalize=False,
                aggr="mean",
            )

    def forward(
        self,
        x: Tensor,
        neighbors: Tensor,
        edge_info: Optional[Tensor] = None,
        batch: Optional[Tensor] = None,
        mask: Optional[Tensor] = None,
    ) -> Tensor:
        """
        Forward pass

        Args:
            x (Tensor): Node features [minibatch_size, ngenes]
            neighbors (Tensor): Neighbor nodes [minibatch_size, ngenes, n_neighbors] or [minibatch_size, ngenes, n_neighbors, 2]
            edge_info (Tensor, optional): Graph connectivity [2, num_edges] if gnn_type != deepset,
                Edge features [num_edges, 1] if gnn_type == deepset,
                or None if gnn_type == deepset and no edge features.
            batch (Tensor, optional): Batch assignment vector [num_nodes]
            mask (Tensor, optional): Mask tensor for the nodes.

        Returns:
            Tensor: Node embeddings [num_nodes, hidden_dim]
        """

        # Standard GNN forward pass
        x = x.unsqueeze(-1)
        neighbors = neighbors.unsqueeze(-1)
        if self.gnn_type == "deepset":
            neighbors = self.input_nn_layer(neighbors).sum(dim=-3)
            x = self.input_self_layer(x)
            x = torch.cat([x, neighbors], dim=-1)
        else:
            x = self.gnn_layer(x, edge_info)
            neighbors = self.gnn_layer(neighbors, edge_info)
            for layer in self.layers:
                x = layer(x, edge_info)
                x = F.relu(x)
                x = F.dropout(x, p=self.dropout, training=self.training)
            # TODO: to finish

        x = self.output_layer(x)
        if mask is not None:
            x = x.masked_fill_(mask.unsqueeze(-1), 0)
        return x



import os
from typing import Any, Dict, List, Optional

import bionty as bt
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scanpy as sc
import seaborn as sns
import torch
from anndata import AnnData, concat
from scdataloader import Collator, Preprocessor
from scdataloader.data import SimpleAnnDataset
from scdataloader.utils import get_descendants, random_str
from scib_metrics.benchmark import Benchmarker
from scipy.stats import spearmanr
from simpler_flash import FlashTransformer
from sklearn.metrics import f1_score
from torch.utils.data import DataLoader
from tqdm import tqdm

FILE_LOC = os.path.dirname(os.path.realpath(__file__))


class Generate:
    def __init__(
        self,
        genelist: List[str],
        batch_size: int = 64,
        embedding_to_use: List[str] = [
            "all",
        ],
    ):
        """
        Embedder a class to embed and annotate cells using a model

        Args:
            genelist (List[str]): The list of genes for which to generate expression data.
            batch_size (int, optional): The size of the batches to be used in the DataLoader. Defaults to 64.
            embedding_to_use (List[str], optional): The list of embeddings to be used for generating expression. Defaults to ["all"].
        """
        self.batch_size = batch_size
        self.embedding_to_use = embedding_to_use
        self.genelist = genelist if genelist is not None else []

    def __call__(self, model: torch.nn.Module, adata: AnnData) -> tuple[AnnData, List[str], np.ndarray, dict]:
        """
        __call__ function to call the embedding

        Args:
            model (torch.nn.Module): The scPRINT model to be used for embedding and annotation.
            adata (AnnData): The annotated data matrix of shape n_obs x n_vars. Rows correspond to cells and columns to genes.

        Raises:
            ValueError: If the model does not have a logger attribute.
            ValueError: If the model does not have a global_step attribute.

        Returns:
            AnnData: The annotated data matrix with embedded cell representations.
            List[str]: List of gene names used in the embedding.
            np.ndarray: The predicted expression values if sample"none".
            dict: Additional metrics and information from the embedding process.
        """
        # one of "all" "sample" "none"
        model.predict_mode = "none"
        model.eval()
        model.on_predict_epoch_start()
        device = model.device.type
        dtype = (
            torch.float16
            if isinstance(model.transformer, FlashTransformer)
            else model.dtype
        )
        if self.embedding_to_use == ["all"]:
            use = [
                i
                for i in adata.obsm.keys()
                if i.startswith("scprint_emb_") and i != "scprint_emb_other"
            ]
        else:
            use = self.embedding_to_use
        res = []
        with (
            torch.no_grad(),
            torch.autocast(device_type=device, dtype=dtype),
        ):
            gene_pos = torch.tensor(
                [model.genes.index(g) for g in self.genelist],
            ).to(device=device)
            gene_pos = gene_pos.unsqueeze(0).repeat_interleave(self.batch_size, 0)
            req_depth = torch.tensor(adata.X.sum(1)).squeeze(-1).to(device=device)

            for batch in tqdm(range(adata.shape[0] // self.batch_size + 1)):
                embeddings = []
                start = batch * self.batch_size
                end = min((batch + 1) * self.batch_size, adata.shape[0])
                for emb in use:
                    embeddings.append(
                        torch.tensor(adata.obsm[emb][start:end]).unsqueeze(1)
                    )
                embeddings = torch.concat(embeddings, dim=1).to(device=device)

                output = model._generate(
                    gene_pos=gene_pos[0 : end - start, :],
                    cell_embs=embeddings,
                    depth_mult=req_depth[start:end],
                    req_depth=req_depth[start:end],
                    metacell_token=None,
                )
                res.append(
                    torch.concat(
                        [
                            output["mean"].detach().cpu().unsqueeze(0),
                            output["disp"].detach().cpu().unsqueeze(0),
                            output["zero_logits"].detach().cpu().unsqueeze(0),
                        ]
                    )
                    if "disp" in output
                    else output["mean"].detach().cpu().unsqueeze(0)
                )
                torch.cuda.empty_cache()
        res = torch.concat(res, dim=1)
        pred_adata = AnnData(
            X=res[0, :, :].numpy(),
            obs=adata.obs.copy(),
            var=pd.DataFrame(
                index=pd.Index(self.genelist),
            ),
            layers=None
            if res.shape[1] == 1
            else {
                "disp": res[1, :, :].numpy(),
                "zero_logits": res[2, :, :].numpy(),
            },
        )
        return pred_adata



from typing import Callable, Dict, List, Optional, Tuple, Union

import torch
from torch import Tensor, nn
from torch.nn import functional as F


class GraphSDEExprDecoder(nn.Module):
    def __init__(self, d_model: int, drift: nn.Module, diffusion: nn.Module):
        """
        Initialize the ExprNeuralSDEDecoder module.

        Args:
            d_model (int): The dimension of the model.
            drift (nn.Module): The drift component of the SDE.
            diffusion (nn.Module): The diffusion component of the SDE.
        """
        super().__init__()
        self.d_model = d_model
        self.drift = drift
        self.diffusion = diffusion

    def forward(self, x: Tensor, dt: float) -> Tensor:
        drift = self.drift(x)
        diffusion = self.diffusion(x)
        dW = torch.randn_like(x) * torch.sqrt(dt)
        return x + drift * dt + diffusion * dW


class ExprDecoder(nn.Module):
    def __init__(
        self,
        d_model: int,
        nfirst_tokens_to_skip: int = 0,
        dropout: float = 0.1,
        zinb: bool = True,
        use_depth: bool = False,
    ):
        """
        ExprDecoder Decoder for the gene expression prediction.

        Will output the mean, variance and zero logits, parameters of a zero inflated negative binomial distribution.

        Args:
            d_model (int): The dimension of the model. This is the size of the input feature vector.
            nfirst_tokens_to_skip (int, optional): The number of initial labels to skip in the sequence. Defaults to 0.
            dropout (float, optional): The dropout rate applied during training to prevent overfitting. Defaults to 0.1.
            zinb (bool, optional): Whether to use a zero inflated negative binomial distribution. Defaults to True.
            use_depth (bool, optional): Whether to use depth as an additional feature. Defaults to False.
        """
        super(ExprDecoder, self).__init__()
        self.fc = nn.Sequential(
            nn.Linear(d_model if not use_depth else d_model + 1, d_model),
            nn.LayerNorm(d_model),
            nn.LeakyReLU(),
            nn.Dropout(dropout),
            nn.Linear(d_model, d_model),
            nn.LayerNorm(d_model),
            nn.LeakyReLU(),
        )
        self.pred_var_zero = nn.Linear(d_model, 3 if zinb else 1)
        self.zinb = zinb

    def forward(
        self, x: Tensor, req_depth: Optional[Tensor] = None
    ) -> Dict[str, Tensor]:
        """x is the output of the transformer, (batch, seq_len, d_model)"""
        # we don't do it on the labels
        if req_depth is not None:
            x = torch.cat(
                [x, req_depth.unsqueeze(1).unsqueeze(-1).expand(-1, x.shape[1], -1)],
                dim=-1,
            )
        x = self.fc(x)
        if self.zinb:
            pred_value, var_value, zero_logits = self.pred_var_zero(x).split(
                1, dim=-1
            )  # (batch, seq_len)
            # The sigmoid function is used to map the zero_logits to a probability between 0 and 1.
            return dict(
                mean=F.softmax(pred_value.squeeze(-1), dim=-1),
                disp=torch.exp(torch.clamp(var_value.squeeze(-1), max=15)),
                zero_logits=zero_logits.squeeze(-1),
            )
        else:
            pred_value = self.pred_var_zero(x)
            return dict(mean=F.softmax(pred_value.squeeze(-1), dim=-1))


class MVCDecoder(nn.Module):
    def __init__(
        self,
        d_model: int,
        arch_style: str = "inner product",
        tot_labels: int = 1,
        query_activation: nn.Module = nn.Sigmoid,
        hidden_activation: nn.Module = nn.PReLU,
        use_depth: bool = False,
        zinb: bool = True,
    ) -> None:
        """
        MVCDecoder Decoder for masked value prediction of cell embeddings.

        Uses gene embeddings with cell embeddings to predict mean, variance, and zero logits
        parameters of a zero-inflated negative binomial distribution.

        Args:
            d_model (int): Dimension of the gene embedding.
            arch_style (str, optional): Architecture style of the decoder. Options:
                "inner product": Uses inner product between cell and gene embeddings,
                "concat query": Concatenates cell and gene embeddings,
                "sum query": Sums cell and gene embeddings.
                Defaults to "inner product".
            tot_labels (int, optional): Total number of labels in the input. Defaults to 1.
            query_activation (nn.Module, optional): Activation function for query vectors. Defaults to nn.Sigmoid.
            hidden_activation (nn.Module, optional): Activation function for hidden layers. Defaults to nn.PReLU.
            use_depth (bool, optional): Whether to use depth as an additional feature. Defaults to False.
            zinb (bool, optional): Whether to use a zero-inflated negative binomial distribution. Defaults to True.
        """
        super(MVCDecoder, self).__init__()
        if arch_style == "inner product":
            self.gene2query = nn.Linear(
                d_model if not use_depth else d_model + 1, d_model
            )
            self.norm = nn.LayerNorm(d_model)
            self.query_activation = query_activation()
            self.pred_var_zero = nn.Linear(
                d_model, d_model * (3 if zinb else 1), bias=False
            )
        elif arch_style == "concat query":
            self.gene2query = nn.Linear(
                d_model if not use_depth else d_model + 1, d_model
            )
            self.query_activation = query_activation()
            self.fc1 = nn.Linear(d_model * (1 + tot_labels), d_model // 2)
            self.hidden_activation = hidden_activation()
            self.fc2 = nn.Linear(d_model // 2, (3 if zinb else 1))
        elif arch_style == "sum query":
            self.gene2query = nn.Linear(
                d_model if not use_depth else d_model + 1, d_model
            )
            self.query_activation = query_activation()
            self.fc1 = nn.Linear(d_model, 64)
            self.hidden_activation = hidden_activation()
            self.fc2 = nn.Linear(64, (3 if zinb else 1))
        else:
            raise ValueError(f"Unknown arch_style: {arch_style}")

        self.arch_style = arch_style
        self.do_detach = arch_style.endswith("detach")
        self.d_model = d_model
        self.zinb = zinb

    def forward(
        self,
        cell_emb: Tensor,
        gene_embs: Tensor,
        req_depth: Optional[Tensor] = None,
    ) -> Union[Tensor, Dict[str, Tensor]]:
        """
        Args:
            cell_emb: Tensor, shape (batch, embsize=d_model)
            gene_embs: Tensor, shape (batch, seq_len, embsize=d_model)
            req_depth: Tensor, shape (batch,), optional depth information.

        Returns:
            Dict[str, Tensor]: A dictionary containing the predicted mean, variance, and zero logits (if zinb is True).
        """
        if req_depth is not None:
            gene_embs = torch.cat(
                [
                    gene_embs,
                    req_depth.unsqueeze(1)
                    .unsqueeze(-1)
                    .expand(-1, gene_embs.shape[1], -1),
                ],
                dim=-1,
            )
        if self.arch_style == "inner product":
            query_vecs = self.query_activation(self.norm(self.gene2query(gene_embs)))
            if self.zinb:
                pred, var, zero_logits = self.pred_var_zero(query_vecs).split(
                    self.d_model, dim=-1
                )
            else:
                pred = self.pred_var_zero(query_vecs)
            cell_emb = cell_emb.unsqueeze(2)
            if self.zinb:
                pred, var, zero_logits = (
                    torch.bmm(pred, cell_emb).squeeze(2),
                    torch.bmm(var, cell_emb).squeeze(2),
                    torch.bmm(zero_logits, cell_emb).squeeze(2),
                )
            else:
                pred = torch.bmm(pred, cell_emb).squeeze(2)
            # zero logits need to based on the cell_emb, because of input exprs
        elif self.arch_style == "concat query":
            query_vecs = self.query_activation(self.gene2query(gene_embs))
            # expand cell_emb to (batch, seq_len, embsize)
            cell_emb = cell_emb.unsqueeze(1).expand(-1, gene_embs.shape[1], -1)

            h = self.hidden_activation(
                self.fc1(torch.cat([cell_emb, query_vecs], dim=2))
            )
            if self.zinb:
                pred, var, zero_logits = self.fc2(h).split(1, dim=-1)
            else:
                pred = self.fc2(h)
        elif self.arch_style == "sum query":
            query_vecs = self.query_activation(self.gene2query(gene_embs))
            cell_emb = cell_emb.unsqueeze(1)

            h = self.hidden_activation(self.fc1(cell_emb + query_vecs))
            if self.zinb:
                pred, var, zero_logits = self.fc2(h).split(1, dim=-1)
            else:
                pred = self.fc2(h)
        if self.zinb:
            return dict(
                mvc_mean=F.softmax(pred, dim=-1),
                mvc_disp=torch.exp(torch.clamp(var, max=15)),
                mvc_zero_logits=zero_logits,
            )
        else:
            return dict(mvc_mean=F.softmax(pred, dim=-1))


class ClsDecoder(nn.Module):
    def __init__(
        self,
        d_model: int,
        n_cls: int,
        layers: List[int] = [256, 128],
        activation: Callable = nn.ReLU,
        dropout: float = 0.1,
    ):
        """
        ClsDecoder Decoder for classification task.

        Args:
            d_model (int): Dimension of the input.
            n_cls (int): Number of classes.
            layers (List[int]): List of hidden layers.
            activation (Callable): Activation function.
            dropout (float): Dropout rate.
        """
        super(ClsDecoder, self).__init__()
        # module List
        layers = [d_model] + layers
        self.decoder = nn.Sequential()
        self.n_cls = n_cls
        for i, l in enumerate(layers[1:]):
            self.decoder.append(nn.Linear(layers[i], l))
            self.decoder.append(nn.LayerNorm(l))
            self.decoder.append(activation())
            self.decoder.append(nn.Dropout(dropout))
        self.out_layer = nn.Linear(layers[-1], n_cls)

    def forward(self, x: Tensor) -> Tensor:
        """
        Args:
            x: Tensor, shape [batch_size, embsize]
        """
        x = self.decoder(x)
        return self.out_layer(x)


class VAEDecoder(nn.Module):
    def __init__(
        self,
        d_model: int,
        layers: List[int] = [64, 64],
        activation: Callable = nn.ReLU,
        dropout: float = 0.1,
        return_latent: bool = False,
    ):
        """
        VAEDecoder for variational autoencoding of cell embeddings.

        Args:
            d_model (int): Input dimension (original embedding size)
            layers (List[int]): List of hidden layer sizes for encoder and decoder
            activation (Callable): Activation function to use
            dropout (float): Dropout rate
            return_latent (bool): Whether to return the latent vectors
        """
        super(VAEDecoder, self).__init__()

        # Encoder layers
        self.return_latent = return_latent
        encoder_layers = [d_model] + layers
        self.encoder = nn.Sequential()
        for i, (in_size, out_size) in enumerate(
            zip(encoder_layers[:-1], encoder_layers[1:])
        ):
            self.encoder.append(nn.Linear(in_size, out_size))
            self.encoder.append(nn.LayerNorm(out_size))
            self.encoder.append(activation())
            self.encoder.append(nn.Dropout(dropout))

        # VAE latent parameters
        self.fc_mu = nn.Linear(encoder_layers[-1], encoder_layers[-1])
        self.fc_var = nn.Linear(encoder_layers[-1], encoder_layers[-1])

        # Decoder layers
        decoder_layers = [encoder_layers[-1]] + list(reversed(layers[:-1])) + [d_model]
        self.decoder = nn.Sequential()
        for i, (in_size, out_size) in enumerate(
            zip(
                decoder_layers[:-1], decoder_layers[1:]
            )  # Changed to include final layer
        ):
            self.decoder.append(nn.Linear(in_size, out_size))
            if (
                i < len(decoder_layers) - 2
            ):  # Don't apply activation/norm to final layer
                self.decoder.append(nn.LayerNorm(out_size))
                self.decoder.append(activation())
                self.decoder.append(nn.Dropout(dropout))

    def reparameterize(self, mu: Tensor, log_var: Tensor) -> Tensor:
        """
        Reparameterization trick to sample from N(mu, var) from N(0,1).

        Args:
            mu (Tensor): Mean of the latent Gaussian
            log_var (Tensor): Log variance of the latent Gaussian

        Returns:
            Tensor: Sampled latent vector
        """
        std = torch.exp(0.5 * log_var)
        eps = torch.randn_like(std)
        return mu + eps * std

    def kl_divergence(self, mu: Tensor, log_var: Tensor) -> Tensor:
        """
        Compute KL divergence between N(mu, var) and N(0, 1).

        Args:
            mu (Tensor): Mean of the latent Gaussian
            log_var (Tensor): Log variance of the latent Gaussian

        Returns:
            Tensor: KL divergence loss
        """
        # KL(N(mu, var) || N(0, 1)) = -0.5 * sum(1 + log(var) - mu^2 - var)
        kl_loss = -0.5 * torch.sum(1 + log_var - mu.pow(2) - log_var.exp(), dim=1)
        return kl_loss.mean()

    def forward(
        self, x: Tensor
    ) -> Union[Tensor, Tuple[Tensor, Tensor, Tensor, Tensor, Tensor]]:
        """
        Forward pass through VAE.

        Args:
            x (Tensor): Input tensor of shape [batch_size, d_model]

        Returns:
            If self.return_latent is True:
                Tuple[Tensor, Tensor, Tensor, Tensor, Tensor]:
                    - reconstructed_x (Tensor): Reconstructed input, shape [batch_size, d_model]
                    - mu (Tensor): Mean of the latent Gaussian, shape [batch_size, latent_dim]
                    - log_var (Tensor): Log variance of the latent Gaussian, shape [batch_size, latent_dim]
                    - kl_loss (Tensor): KL divergence loss (scalar tensor)
            Else:
                Tensor: reconstructed_x of shape [batch_size, d_model]
        """
        # Encode
        encoded = self.encoder(x)

        # Get latent parameters
        mu = self.fc_mu(encoded)
        log_var = self.fc_var(encoded)
        log_var = torch.clamp(log_var, min=-10)

        # Sample latent vector
        kl_loss = self.kl_divergence(mu, log_var)
        # free_bits = 2.0  # per latent dim
        # kl_loss = torch.clamp(kl_loss / mu.size(-1), min=free_bits) * mu.size(-1)
        z = self.reparameterize(mu, log_var)

        # Decode
        decoded = self.decoder(z)

        if self.return_latent:
            return decoded, mu, log_var, encoded, kl_loss
        return decoded, kl_loss



import os
import sys

import numpy as np
import pandas as pd
import scanpy as sc
import torch
from anndata import AnnData
from scdataloader import Collator
from scdataloader.data import SimpleAnnDataset
from simpler_flash import FlashTransformer
from torch.cuda.amp import autocast
from torch.utils.data import DataLoader


class GeneEmbeddingExtractor:
    def __init__(
        self,
        genelist: list[str],
        batch_size: int = 64,
        num_workers: int = 8,
        save_every: int = 4_000,
        average: bool = False,
        save_dir: str = None,
        use_knn: bool = False,
    ):
        """
        Args:
            genelist (list[str]): List of genes to restrict to.
            batch_size (int): Batch size for the DataLoader. Defaults to 64.
            num_workers (int): Number of workers for DataLoader. Defaults to 8.
            save_every (int): Save embeddings every `save_every` batches. Defaults to 4000.
            average (bool): Whether to average embeddings across all cells. Defaults to False.
            save_dir (str): Directory to save embeddings. If None, embeddings are not saved. Defaults to None.
            use_knn (bool): Whether to use k-nearest neighbors information. Defaults to False.

        """
        self.genelist = genelist
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.save_every = save_every
        self.average = average
        self.save_dir = save_dir
        self.use_knn = use_knn

    def __call__(self, model, adata):
        model.eval()
        model.pred_log_adata = False
        model.predict_mode = "none"

        # Determine which genes to use
        prevl = len(self.genelist)
        gene_list = [g for g in self.genelist if g in model.genes]
        print(
            "Using {:.2f}% of the genes in the gene list".format(
                len(self.genelist) * 100 / prevl
            )
        )
        if len(gene_list) == 0:
            raise ValueError("No overlap between provided gene_list and model.genes")

        # Set up dataset and dataloader
        # If needed, ensure adata.obs contains 'organism_ontology_term_id' or adapt Collator arguments
        if "organism_ontology_term_id" not in adata.obs:
            # Assign a default organism if needed
            adata.obs["organism_ontology_term_id"] = (
                "NCBITaxon:9606"  # or your relevant organism ID
            )

        adataset = SimpleAnnDataset(adata, obs_to_output=["organism_ontology_term_id"])
        col = Collator(
            organisms=model.organisms,
            valid_genes=model.genes,
            max_len=0,
            how="some",
            genelist=gene_list,
            n_bins=model.n_input_bins if model.expr_emb_style == "binned" else 0,
        )
        dataloader = DataLoader(
            adataset,
            collate_fn=col,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=False,
        )

        all_embeddings = []
        expr = []
        model.on_predict_epoch_start()
        count = 0
        full_embeddings = None
        dtype = (
            torch.float16
            if isinstance(model.transformer, FlashTransformer)
            else torch.float32
        )
        # Use autocast to ensure half precision if required by the model
        with (
            torch.no_grad(),
            torch.autocast(device_type=model.device.type, dtype=dtype),
        ):
            for batch in dataloader:
                gene_pos, expression, depth = (
                    batch["genes"].to(model.device),
                    batch["x"].to(model.device),
                    batch["depth"].to(model.device),
                )

                # Run encode_only to get transformer outputs
                output = model(
                    gene_pos=gene_pos,
                    expression=expression,
                    req_depth=depth,
                    depth_mult=expression.sum(1),
                    get_gene_emb=True,
                )
                # transformer_output shape: (B, cell_embs_count + num_genes, d_model)
                # Extract gene embeddings:
                gene_embeddings = output["gene_embedding"]
                # shape: (B, num_genes, d_model)
                if self.average:
                    if full_embeddings is None:
                        # Average the gene embeddings across the batch
                        full_embeddings = gene_embeddings.mean(dim=0, keepdim=True)
                    else:
                        full_embeddings += gene_embeddings.mean(dim=0, keepdim=True)
                    count += 1
                else:
                    all_embeddings.append(gene_embeddings.cpu().numpy())
                expr.append(batch["x"].numpy())
                if (
                    len(all_embeddings) * self.batch_size
                ) >= self.save_every and not self.average:
                    # Save embeddings to disk or process them as needed
                    ad = AnnData(
                        X=np.concatenate(expr, axis=0),
                        var=pd.DataFrame(index=gene_list),
                        varp=np.concatenate(all_embeddings, axis=0),
                    )
                    if self.save_dir is None:
                        print(
                            "reached max len, need to save embeddings but did not specify save_dir"
                        )
                        print("using default save_dir: ./data/")
                        self.save_dir = "./data/"
                    if not os.path.exists(self.save_dir):
                        os.makedirs(self.save_dir)
                    ad.write_h5ad(
                        os.path.join(self.save_dir, f"embeddings_{count}.h5ad")
                    )
                    count += 1
                    del ad
                    all_embeddings = []
                    expr = []
                del output
                torch.cuda.empty_cache()
        ad = AnnData(
            X=np.concatenate(expr, axis=0),
            var=pd.DataFrame(index=gene_list),
            uns=(
                {"all_embeddings": np.concatenate(all_embeddings, axis=0)}
                if not self.average
                else None
            ),
            varm=full_embeddings.cpu().numpy() / count if self.average else None,
        )
        if self.save_dir is not None:
            ad.write_h5ad(os.path.join(self.save_dir, f"embeddings_{count}.h5ad"))
            print("save {} embeddings files under {}".format(count, self.save_dir))
        return ad



import gc
import os.path
from typing import Any, List, Optional

import joblib
import networkx as nx
import numpy as np
import scanpy as sc
import scipy.sparse
import seaborn as sns
import torch
import umap
from anndata import AnnData
from anndata.utils import make_index_unique
from bengrn import BenGRN, get_perturb_gt, get_sroy_gt
from bengrn.base import train_classifier

# from bengrn.GeneRNIB_reg2 import run_gene_rnib, NORMAN, OP, ADAMSON
from grnndata import GRNAnnData, from_anndata
from grnndata import utils as grnutils
from matplotlib import pyplot as plt
from scdataloader import Collator, Preprocessor
from scdataloader.data import SimpleAnnDataset
from scdataloader.utils import load_genes
from simpler_flash import FlashTransformer
from torch.utils.data import DataLoader
from tqdm import tqdm

from scprint2.utils.sinkhorn import SinkhornDistance

from .tmfg import tmfg

FILEDIR = os.path.dirname(os.path.realpath(__file__))


class GNInfer:
    def __init__(
        self,
        batch_size: int = 64,
        num_workers: int = 8,
        drop_unexpressed: bool = True,
        num_genes: int = 3000,
        max_cells: int = 0,
        cell_type_col: str = "cell_type",
        how: str = "most var within",  # random expr, most var within, most var across, some
        genelist: Optional[List[str]] = None,
        layer: Optional[List[int]] = None,
        preprocess: str = "softmax",  # sinkhorn, softmax, none
        head_agg: str = "mean",  # mean, sum, none, mean_full
        filtration: str = "thresh",  # thresh, top-k, mst, known, none
        k: int = 10,
        known_grn: Optional[Any] = None,
        precomp_attn: bool = True,
        symmetrize: bool = False,
        loc: str = "./",
        use_knn: bool = True,
    ):
        """
        GNInfer a class to infer gene regulatory networks from a dataset using a scPRINT model.

        Args:
            batch_size (int, optional): Batch size for processing. Defaults to 64.
            num_workers (int, optional): Number of workers for data loading. Defaults to 8.
            drop_unexpressed (bool, optional): Whether to drop unexpressed genes. Defaults to True.
                In this context, genes that have no expression in the dataset are dropped.
            num_genes (int, optional): Number of genes to consider. Defaults to 3000.
            max_cells (int, optional): Maximum number of cells to consider. Defaults to 0.
                if less than total number of cells, only the top `max_cells` cells with the most counts will be considered.
            cell_type_col (str, optional): Column name for cell type information. Defaults to "cell_type".
            how (str, optional): Method to select genes. Options are "most var", "random expr", "some". Defaults to "most var".
                - "most var across": select the most variable genes across all cell types
                - "most var within": select the most variable genes within a cell type
                - "random expr": select random expressed genes
                - "some": select a subset of genes defined in genelist
                - "most expr": select the most expressed genes in the cell type
            genelist (list, optional): List of genes to consider. Defaults to an empty list.
            layer (Optional[List[int]], optional): List of layers to use for the inference. Defaults to None.
            preprocess (str, optional): Preprocessing method. Options are "softmax", "sinkhorn", "none". Defaults to "softmax".
            head_agg (str, optional): Aggregation method for heads. Options are "mean_full", "mean", "sum", "none". Defaults to "mean".
            filtration (str, optional): Filtration method for the adjacency matrix. Options are "thresh", "top-k", "mst", "known", "none". Defaults to "thresh".
            k (int, optional): Number of top connections to keep if filtration is "top-k". Defaults to 10.
            known_grn (optional): Known gene regulatory network to use as a reference. Defaults to None.
                - We will only keep the genes that are present in the known GRN.
            precomp_attn (bool, optional): Whether to let the model precompute attn or do it at the end.
                This takes more memory but the model can compute mean over the attention matrices instead
                of over the qs and ks then taking the product.
                It is required for mean_full head_agg. Defaults to False.
            symmetrize (bool, optional): Whether to GRN. Defaults to False.
            loc (str, optional): Location to save results. Defaults to "./".
            use_knn (bool, optional): Whether to use k-nearest neighbors information. Defaults to True.
        """
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.layer = layer
        self.loc = loc
        self.how = how
        assert self.how in [
            "most var within",
            "most var across",
            "random expr",
            "some",
            "most expr",
        ], "how must be one of 'most var within', 'most var across', 'random expr', 'some', 'most expr'"
        self.num_genes = num_genes if self.how != "some" else len(self.genelist)
        self.preprocess = preprocess
        self.cell_type_col = cell_type_col
        self.filtration = filtration
        self.genelist = genelist if genelist is not None else []
        self.k = k
        self.symmetrize = symmetrize
        self.known_grn = known_grn
        self.head_agg = head_agg
        self.max_cells = max_cells
        self.curr_genes = None
        self.drop_unexpressed = drop_unexpressed
        self.use_knn = use_knn
        if self.filtration != "none" and self.head_agg == "none":
            raise ValueError("filtration must be 'none' when head_agg is 'none'")

    def __call__(
        self, model: torch.nn.Module, adata: AnnData, cell_type=None
    ) -> tuple[AnnData, np.ndarray]:
        """
        __call__ runs the method

        Args:
            model (torch.nn.Module): The model to be used for generating the network
            adata (AnnData): Annotated data matrix of shape `n_obs` × `n_vars`. `n_obs` is the number of cells and `n_vars` is the number of genes.
            cell_type (str, optional): Specific cell type to filter the data. Defaults to None.

        Returns:
            AnnData: Annotated data matrix with predictions and annotations.
            np.ndarray: Filtered adjacency matrix.
        """
        # Add at least the organism you are working with
        if self.layer is None:
            self.layer = list(range(model.nlayers))
        self.n_cell_embs = model.attn.additional_tokens
        subadata = self.predict(model, adata, self.layer, cell_type)
        adjacencies = self.aggregate(model)
        model.attn.data = None
        if self.head_agg == "none":
            return self.save(
                adjacencies[self.n_cell_embs :, self.n_cell_embs :, :],
                subadata,
            )
        else:
            return self.save(
                self.filter(adjacencies)[self.n_cell_embs :, self.n_cell_embs :],
                subadata,
                loc=self.loc,
            )

    def predict(self, model, adata, layer, cell_type=None):
        """
        part to predict the qks or attns matrices from the adata with the model
        """
        self.curr_genes = None
        model.pred_log_adata = False
        if cell_type is not None:
            subadata = adata[adata.obs[self.cell_type_col] == cell_type].copy()
        else:
            subadata = adata.copy()
        if self.how == "most var within":
            try:
                sc.pp.highly_variable_genes(
                    subadata, flavor="seurat_v3", n_top_genes=self.num_genes
                )
            except ValueError:
                sc.pp.highly_variable_genes(
                    subadata,
                    flavor="seurat_v3",
                    n_top_genes=self.num_genes,
                    span=0.6,
                )
            self.curr_genes = (
                subadata.var.index[subadata.var.highly_variable].tolist()
                + self.genelist
            )
            print(
                "number of expressed genes in this cell type: "
                + str((subadata.X.sum(0) > 1).sum())
            )
        elif self.how == "most var across" and cell_type is not None:
            adata.raw = adata
            sc.tl.rank_genes_groups(
                adata,
                mask_var=adata.var.index.isin(model.genes),
                groupby=self.cell_type_col,
                groups=[cell_type],
            )
            diff_expr_genes = adata.uns["rank_genes_groups"]["names"][cell_type]
            diff_expr_genes = [gene for gene in diff_expr_genes if gene in model.genes]
            self.curr_genes = diff_expr_genes[: self.num_genes] + self.genelist
            self.curr_genes.sort()
        elif self.how == "random expr":
            self.curr_genes = model.genes
            # raise ValueError("cannot do it yet")
            pass
        elif self.how == "some" and len(self.genelist) > 0:
            self.curr_genes = self.genelist
        elif self.how == "most expr":
            self.curr_genes = adata.var.index[
                adata.X.sum(0).A1.argsort()[::-1]
            ].tolist()[: self.num_genes]
        else:
            raise ValueError("something wrong with your inputs")
        if self.drop_unexpressed:
            expr = subadata.var[(subadata.X.sum(0) > 0).tolist()[0]].index.tolist()
            self.curr_genes = [i for i in self.curr_genes if i in expr]
        # Order cells by total count
        cell_sums = subadata.X.sum(axis=1)
        order = np.argsort(
            -cell_sums.A1 if scipy.sparse.issparse(subadata.X) else -cell_sums
        )
        subadata = subadata[order].copy()
        subadata = subadata[: self.max_cells] if self.max_cells else subadata
        if len(subadata) == 0:
            raise ValueError("no cells in the dataset")
        adataset = SimpleAnnDataset(
            subadata,
            obs_to_output=["organism_ontology_term_id"],
            get_knn_cells=model.expr_emb_style == "metacell" and self.use_knn,
        )
        col = Collator(
            organisms=model.organisms,
            valid_genes=model.genes,
            max_len=self.num_genes if self.how == "random expr" else 0,
            how="some" if self.how != "random expr" else "random expr",
            genelist=self.curr_genes if self.how != "random expr" else [],
            n_bins=model.n_input_bins if model.expr_emb_style == "binned" else 0,
        )
        dataloader = DataLoader(
            adataset,
            collate_fn=col,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=False,
        )
        model.attn.precomp_attn = self.head_agg == "mean_full"
        if self.num_genes > 10_000 and model.attn.precomp_attn:
            raise ValueError("need less genes for a non-shared-qk version")
        prevplot = model.doplot

        model.doplot = False
        model.on_predict_epoch_start()
        model.eval()
        model.attn.data = None
        # reparametrize the attn process

        if model.transformer.attn_type == "hyper":
            self.curr_genes = [i for i in model.genes if i in self.curr_genes]
            num = (1 if model.use_metacell_token else 0) + (
                (len(model.classes) + 1) if not model.cell_transformer else 0
            )
            if (len(self.curr_genes) + num) % 128 != 0:
                self.curr_genes = self.curr_genes[
                    : (len(self.curr_genes) // 128 * 128) - num
                ]
        if self.how != "random expr":
            if model.attn.precomp_attn:
                model.attn.gene_dim = len(set(self.curr_genes) & set(model.genes))
                model.attn.apply_softmax = self.preprocess == "softmax"
            else:
                if subadata.obs["organism_ontology_term_id"].unique().shape[0] > 1:
                    raise ValueError(
                        "only one organism at a time is supported for precomp_attn"
                    )
                n = False
                for i, k in col.start_idx.items():
                    if n:
                        model.attn.gene_dim = k - model.attn.speciesloc
                        break
                    if i == subadata.obs["organism_ontology_term_id"].unique()[0]:
                        model.attn.speciesloc = k
                        n = True
        elif not model.attn.precomp_attn:
            raise ValueError(
                "full attention (i.e. precomp_attn=True) is not supported for random expr"
            )
        device = model.device.type
        dtype = (
            torch.float16
            if isinstance(model.transformer, FlashTransformer)
            else model.dtype
        )
        with torch.no_grad(), torch.autocast(device_type=device, dtype=dtype):
            for batch in tqdm(dataloader):
                gene_pos, expression, depth = (
                    batch["genes"].to(device),
                    batch["x"].to(device),
                    batch["depth"].to(device),
                )
                model._predict(
                    gene_pos,
                    expression,
                    depth,
                    knn_cells=(
                        batch["knn_cells"].to(device)
                        if model.expr_emb_style == "metacell" and self.use_knn
                        else None
                    ),
                    knn_cells_info=(
                        batch["knn_cells_info"].to(device)
                        if model.expr_emb_style == "metacell" and self.use_knn
                        else None
                    ),
                    keep_output=False,
                    get_attention_layer=layer if type(layer) is list else [layer],
                )
                torch.cuda.empty_cache()
        model.doplot = prevplot
        return subadata

    def aggregate(self, model):
        """
        part to aggregate the qks and compute the attns
        or to aggregate the attns
        or do nothing if already done
        """
        attn, genes = model.attn.get(), model.genes
        if model.attn.precomp_attn:
            self.curr_genes = [i for i in genes if i in self.curr_genes]
            return attn.detach().cpu().numpy()
        if self.how == "random expr" and self.drop_unexpressed:
            keep = np.array(
                [1] * self.n_cell_embs + [i in self.curr_genes for i in genes],
                dtype=bool,
            )
            attn = attn[:, keep, :, :, :]
        badloc = torch.isnan(attn.sum((0, 2, 3, 4)))
        attn = attn[:, ~badloc, :, :, :]
        badloc = badloc.detach().cpu().numpy()
        self.curr_genes = (
            np.array(self.curr_genes)[~badloc[self.n_cell_embs :]]
            if self.how == "random expr"
            else [i for i in genes if i in self.curr_genes]
        )
        # attn = attn[:, :, 0, :, :].permute(0, 2, 1, 3) @ attn[:, :, 1, :, :].permute(
        #    0, 2, 3, 1
        # )
        attns = None
        Qs = (
            attn[:, :, 0, :, :]
            .permute(0, 2, 1, 3)
            .reshape(-1, attn.shape[1], attn.shape[-1])
        )
        Ks = (
            attn[:, :, 1, :, :]
            .permute(0, 2, 1, 3)
            .reshape(-1, attn.shape[1], attn.shape[-1])
        )
        for i in range(Qs.shape[0]):
            attn = Qs[i] @ Ks[i].T
            # return attn

            if self.preprocess == "sinkhorn":
                scale = Qs.shape[-1] ** -0.5
                attn = attn * scale
                if attn.numel() > 100_000_000:
                    raise ValueError("you can't sinkhorn such a large matrix")
                sink = SinkhornDistance(0.1, max_iter=200)
                attn = sink(attn)[0]
                attn = attn * Qs.shape[-1]
            elif self.preprocess == "softmax":
                scale = Qs.shape[-1] ** -0.5
                attn = attn * scale
                attn = torch.nn.functional.softmax(attn, dim=-1)
            elif self.preprocess == "softpick":
                attn = softpick(attn)
            elif self.preprocess == "none":
                pass
            else:
                raise ValueError(
                    "preprocess must be one of 'sinkhorn', 'softmax', 'none'"
                )
            if self.symmetrize:
                attn = (attn + attn.T) / 2
            if self.head_agg == "mean":
                attns = attn + (attns if attns is not None else 0)
            elif self.head_agg == "max":
                attns = torch.max(attn, attns) if attns is not None else attn
            elif self.head_agg == "none":
                attn = attn.reshape(attn.shape[0], attn.shape[1], 1)
                if attns is not None:
                    attns = torch.cat((attns, attn.detach().cpu()), dim=2)
                else:
                    attns = attn.detach().cpu()
            else:
                raise ValueError(
                    "head_agg must be one of 'mean', 'mean_full', 'max' or 'none'"
                )
        if self.head_agg == "mean":
            attns = attns / Qs.shape[0]
        return (
            attns.detach().cpu().numpy() if self.head_agg != "none" else attns.numpy()
        )

    def filter(self, adj, gt=None):
        """
        part to filter the attn matrix given user inputs
        """
        if self.filtration == "thresh":
            adj[adj < (1 / adj.shape[-1])] = 0
            res = (adj != 0).sum()
            if res / adj.shape[0] ** 2 < 0.01:
                adj = scipy.sparse.csr_matrix(adj)
        elif self.filtration == "none":
            pass
        elif self.filtration == "top-k":
            args = np.argsort(adj)
            adj[np.arange(adj.shape[0])[:, None], args[:, : -self.k]] = 0
            adj = scipy.sparse.csr_matrix(adj)
        elif self.filtration == "known" and gt is not None:
            gt = gt.reindex(sorted(gt.columns), axis=1)
            gt = gt.reindex(sorted(gt.columns), axis=0)
            gt = gt[gt.index.isin(self.curr_genes)].iloc[
                :, gt.columns.isin(self.curr_genes)
            ]

            loc = np.isin(self.curr_genes, gt.index)
            self.curr_genes = np.array(self.curr_genes)[loc]
            adj = adj[self.n_cell_embs :, self.n_cell_embs :][loc][:, loc]
            adj[gt.values != 1] = 0
            adj = scipy.sparse.csr_matrix(adj)
        elif self.filtration == "tmfg":
            adj = nx.to_scipy_sparse_array(tmfg(adj))
        elif self.filtration == "mst":
            pass
        else:
            raise ValueError("filtration must be one of 'thresh', 'none' or 'top-k'")
        res = (adj != 0).sum() if self.filtration != "none" else adj.shape[0] ** 2
        print(f"avg link count: {res}, sparsity: {res / adj.shape[0] ** 2}")
        return adj

    def save(self, grn, subadata, loc=""):
        grn = GRNAnnData(
            subadata[:, subadata.var.index.isin(self.curr_genes)].copy(), grn=grn
        )
        # grn = grn[:, (grn.X != 0).sum(0) > (self.max_cells / 32)]
        grn.var["TFs"] = [
            True if i in grnutils.TF else False for i in grn.var["symbol"]
        ]
        grn.uns["grn_scprint_params"] = {
            "filtration": self.filtration,
            "how": self.how,
            "preprocess": self.preprocess,
            "head_agg": self.head_agg,
        }
        if loc != "":
            grn.write_h5ad(loc + "grn_fromscprint.h5ad")
            return from_anndata(grn)
        else:
            return grn


def default_benchmark(
    model: Any,
    default_dataset: str = "sroy",
    cell_types: List[str] = [],
    maxlayers: int = 16,
    maxgenes: int = 5000,
    batch_size: int = 32,
    maxcells: int = 1024,
) -> dict:
    """
    default_benchmark function to run the default scPRINT GRN benchmark

    Args:
        model (Any): The scPRINT model to be used for the benchmark.
        default_dataset (str, optional): The default dataset to use for benchmarking. Defaults to "sroy".
        cell_types (List[str], optional): List of cell types to include in the benchmark. Defaults to [].
        maxlayers (int, optional): Maximum number of layers to use from the model. Defaults to 16.
        maxgenes (int, optional): Maximum number of genes to consider. Defaults to 5000.
        batch_size (int, optional): Batch size for processing. Defaults to 32.
        maxcells (int, optional): Maximum number of cells to consider. Defaults to 1024.

    Returns:
        dict: A dictionary containing the benchmark metrics.
    """
    metrics = {}
    layers = list(range(model.nlayers))[max(0, model.nlayers - maxlayers) :]
    clf_omni = None
    if default_dataset == "sroy":
        preprocessor = Preprocessor(
            is_symbol=True,
            force_preprocess=True,
            skip_validate=True,
            do_postp=model.expr_emb_style == "metacell",
            min_valid_genes_id=5000,
            min_dataset_size=64,
            keepdata=True,
        )
        clf_self = None
        todo = [
            ("han", "human", "full"),
            ("mine", "human", "full"),
            ("han", "human", "chip"),
            ("han", "human", "ko"),
            ("tran", "mouse", "full"),
            ("zhao", "mouse", "full"),
            ("tran", "mouse", "chip"),
            ("tran", "mouse", "ko"),
        ]
        for da, spe, gt in todo:
            if gt != "full":
                continue
            if "NCBITaxon:10090" not in model.organisms and spe == "mouse":
                continue
            print(da + "_" + gt)
            preadata = get_sroy_gt(get=da, species=spe, gt=gt)
            adata = preprocessor(preadata.copy())
            if model.expr_emb_style == "metacell":
                sc.pp.neighbors(adata, use_rep="X_pca")
            grn_inferer = GNInfer(
                layer=layers,
                how="most var within",
                preprocess=(
                    "softpick"
                    if model.attention in ["softpick", "softpick-flash"]
                    else "softmax"
                ),
                head_agg="none",
                filtration="none",
                num_genes=maxgenes,
                num_workers=8,
                max_cells=maxcells,
                batch_size=batch_size,
            )
            grn = grn_inferer(model, adata)
            grn.varp["all"] = grn.varp["GRN"]
            grn.var["ensembl_id"] = grn.var.index
            grn.var["symbol"] = make_index_unique(grn.var["symbol"].astype(str))
            grn.var.index = grn.var["symbol"]
            grn.varp["GRN"] = grn.varp["all"].mean(-1).T
            metrics["mean_" + da + "_" + gt] = BenGRN(
                grn, do_auc=True, doplot=False
            ).compare_to(other=preadata)
            grn.varp["GRN"] = grn.varp["GRN"].T
            if spe == "human":
                metrics["mean_" + da + "_" + gt + "_base"] = BenGRN(
                    grn, do_auc=True, doplot=False
                ).scprint_benchmark()

            ## OMNI
            if clf_omni is None:
                grn.varp["GRN"] = grn.varp["all"]
                _, m, clf_omni = train_classifier(
                    grn,
                    C=1,
                    train_size=0.9,
                    class_weight={1: 800, 0: 1},
                    shuffle=True,
                    return_full=False,
                )
                joblib.dump(clf_omni, "clf_omni.pkl")
                metrics["omni_classifier"] = m
            coef = clf_omni.coef_[0] if clf_omni.coef_.shape[0] == 1 else clf_omni.coef_
            grn.varp["GRN"] = grn.varp["all"][:, :, coef > 0].mean(-1)
            if spe == "human":
                metrics["omni_" + da + "_" + gt + "_base"] = BenGRN(
                    grn, do_auc=True, doplot=True
                ).scprint_benchmark()
            grn.varp["GRN"] = grn.varp["GRN"].T
            metrics["omni_" + da + "_" + gt] = BenGRN(
                grn, do_auc=True, doplot=False
            ).compare_to(other=preadata)

            ## SELF
            if clf_self is None:
                grn.varp["GRN"] = np.transpose(grn.varp["all"], (1, 0, 2))
                _, m, clf_self = train_classifier(
                    grn,
                    other=preadata,
                    C=1,
                    train_size=0.5,
                    class_weight={1: 40, 0: 1},
                    shuffle=False,
                    return_full=False,
                )
                metrics["self_classifier"] = m
            coef = clf_self.coef_[0] if clf_self.coef_.shape[0] == 1 else clf_self.coef_
            grn.varp["GRN"] = grn.varp["all"][:, :, coef > 0].mean(-1).T
            metrics["self_" + da + "_" + gt] = BenGRN(
                grn, do_auc=True, doplot=False
            ).compare_to(other=preadata)
            if spe == "human":
                grn.varp["GRN"] = grn.varp["GRN"].T
                metrics["self_" + da + "_" + gt + "_base"] = BenGRN(
                    grn, do_auc=True, doplot=True
                ).scprint_benchmark()

            ## chip / ko
            if (da, spe, "chip") in todo:
                preadata = get_sroy_gt(get=da, species=spe, gt="chip")
                grn.varp["GRN"] = grn.varp["all"].mean(-1).T
                metrics["mean_" + da + "_" + "chip"] = BenGRN(
                    grn, do_auc=True, doplot=False
                ).compare_to(other=preadata)
                grn.varp["GRN"] = grn.varp["all"][:, :, coef > 0].mean(-1).T
                metrics["omni_" + da + "_" + "chip"] = BenGRN(
                    grn, do_auc=True, doplot=False
                ).compare_to(other=preadata)

                grn.varp["GRN"] = grn.varp["all"][:, :, coef > 0].mean(-1).T
                metrics["self_" + da + "_" + "chip"] = BenGRN(
                    grn, do_auc=True, doplot=False
                ).compare_to(other=preadata)
            if (da, spe, "ko") in todo:
                preadata = get_sroy_gt(get=da, species=spe, gt="ko")
                grn.varp["GRN"] = grn.varp["all"].mean(-1).T
                metrics["mean_" + da + "_" + "ko"] = BenGRN(
                    grn, do_auc=True, doplot=False
                ).compare_to(other=preadata)
                grn.varp["GRN"] = grn.varp["all"][:, :, coef > 0].mean(-1).T
                metrics["omni_" + da + "_" + "ko"] = BenGRN(
                    grn, do_auc=True, doplot=False
                ).compare_to(other=preadata)
                grn.varp["GRN"] = grn.varp["all"][:, :, coef > 0].mean(-1).T
                metrics["self_" + da + "_" + "ko"] = BenGRN(
                    grn, do_auc=True, doplot=False
                ).compare_to(other=preadata)
            del grn
    elif default_dataset == "gwps":
        adata = get_perturb_gt()
        preprocessor = Preprocessor(
            force_preprocess=True,
            keepdata=True,
            skip_validate=True,
            do_postp=model.expr_emb_style == "metacell",
            min_valid_genes_id=maxgenes,
            min_dataset_size=64,
        )
        nadata = preprocessor(adata.copy())
        if model.expr_emb_style == "metacell":
            sc.pp.neighbors(nadata, use_rep="X_pca")
        nadata.var["isTF"] = False
        nadata.var.loc[nadata.var.gene_name.isin(grnutils.TF), "isTF"] = True
        nadata.var["isTF"].sum()
        grn_inferer = GNInfer(
            layer=layers,
            how="most var within",
            preprocess=(
                "softpick"
                if model.attention in ["softpick", "softpick-flash"]
                else "softmax"
            ),
            head_agg="none",
            filtration="none",
            num_genes=maxgenes,
            max_cells=maxcells,
            num_workers=8,
            batch_size=batch_size,
        )
        grn = grn_inferer(model, nadata)
        del nadata
        grn.varp["all"] = grn.varp["GRN"]

        grn.varp["GRN"] = grn.varp["all"].mean(-1).T
        metrics["mean"] = BenGRN(grn, do_auc=True, doplot=False).compare_to(other=adata)
        grn.var["ensembl_id"] = grn.var.index
        grn.var.index = grn.var["symbol"]
        grn.varp["GRN"] = grn.varp["all"].mean(-1)
        metrics["mean_base"] = BenGRN(
            grn, do_auc=True, doplot=False
        ).scprint_benchmark()

        grn.varp["GRN"] = grn.varp["all"]
        grn.var.index = grn.var["ensembl_id"]
        _, m, clf_omni = train_classifier(
            grn,
            C=1,
            train_size=0.9,
            class_weight={1: 800, 0: 1},
            shuffle=True,
            doplot=False,
            return_full=False,
            use_col="gene_name",
        )
        coef = clf_omni.coef_[0] if clf_omni.coef_.shape[0] == 1 else clf_omni.coef_
        grn.varp["GRN"] = grn.varp["all"][:, :, coef > 0].mean(-1).T
        metrics["omni"] = BenGRN(grn, do_auc=True, doplot=False).compare_to(other=adata)
        metrics["omni_classifier"] = m
        grn.var.index = grn.var["symbol"]
        grn.varp["GRN"] = grn.varp["GRN"].T
        metrics["omni_base"] = BenGRN(
            grn, do_auc=True, doplot=False
        ).scprint_benchmark()
        grn.varp["GRN"] = np.transpose(grn.varp["all"], (1, 0, 2))
        grn.var.index = grn.var["ensembl_id"]
        _, m, clf_self = train_classifier(
            grn,
            other=adata,
            C=1,
            train_size=0.5,
            class_weight={1: 40, 0: 1},
            doplot=False,
            shuffle=False,
            return_full=False,
            use_col="ensembl_id",
        )
        coef = clf_self.coef_[0] if clf_self.coef_.shape[0] == 1 else clf_self.coef_
        grn.varp["GRN"] = grn.varp["all"][:, :, coef > 0].mean(-1).T
        metrics["self"] = BenGRN(grn, do_auc=True, doplot=False).compare_to(other=adata)
        metrics["self_classifier"] = m
        grn.var.index = grn.var["symbol"]
        grn.varp["GRN"] = grn.varp["GRN"].T
        metrics["self_base"] = BenGRN(
            grn, do_auc=True, doplot=False
        ).scprint_benchmark()
    elif default_dataset == "genernib":
        raise ValueError("Not implemented")
        # for adata in [NORMAN, OP, ADAMSON]:
        #   adata = sc.read_h5ad(adata)
        #   adata.obs["organism_ontology_term_id"] = "NCBITaxon:9606"
        #   preprocessor = Preprocessor(
        #       force_preprocess=False,
        #       skip_validate=True,
        #       drop_non_primary=False,
        #       do_postp=False,
        #       min_valid_genes_id=1000,
        #       min_dataset_size=64,
        #       keepdata=True,
        #       is_symbol=True,
        #       use_raw=False,
        #   )
        #   adata = preprocessor(adata.copy())
        #   run_gene_rnib(
        #      adata=adata,
        #      model=model,
        #      layer=layers,
        #      how="most var within",
        #      preprocess="softmax",
        #   )
        #   grn_inferer = GNInfer(
        #      how="most var across",
        #      preprocess="softmax",
        #      head_agg="mean",
        #      filtration="none",
        #      forward_mode="none",
        #      num_genes=3_000,
        #      max_cells=3000,
        #      batch_size=10,
        #      cell_type_col="perturbation",
        #      layer=list(range(model.nlayers))[:],
        #   )
        # grn = grn_inferer(model, adata, cell_type="ctrl")
        # grn.var.index = make_index_unique(grn.var["symbol"].astype(str))

    else:
        # max_genes=4000
        if default_dataset.startswith("https://"):
            adata = sc.read(
                FILEDIR + "/../../data/" + default_dataset.split("/")[-1],
                backup_url=default_dataset,
            )
        else:
            adata = sc.read_h5ad(default_dataset)
        if default_dataset.split("/")[-1] in ["yBCKp6HmXuHa0cZptMo7.h5ad"]:
            use_layer = "counts"
            is_symbol = True
        else:
            use_layer = None
            is_symbol = False

        preprocessor = Preprocessor(
            use_layer=use_layer,
            is_symbol=is_symbol,
            force_preprocess=True,
            skip_validate=True,
            do_postp=model.expr_emb_style == "metacell",
            drop_non_primary=False,
        )
        adata = preprocessor(adata.copy())

        adata.var["isTF"] = False
        adata.var.loc[adata.var.symbol.isin(grnutils.TF), "isTF"] = True
        if model.expr_emb_style == "metacell":
            if "X_pca" not in adata.obsm:
                sc.pp.pca(adata, n_comps=50)
            sc.pp.neighbors(adata, use_rep="X_pca")
        for celltype in list(adata.obs["cell_type"].unique())[:14]:
            # print(celltype)
            # grn_inferer = GNInfer(
            #    layer=layers,
            #    how="random expr",
            #    preprocess="softmax",
            #    head_agg="max",
            #    filtration="none",
            #    num_workers=8,
            #    num_genes=2200,
            #    max_cells=maxcells,
            #    batch_size=batch_size,
            # )
            #
            # grn = grn_inferer(model, adata[adata.X.sum(1) > 500], cell_type=celltype)
            # grn.var.index = make_index_unique(grn.var["symbol"].astype(str))
            # metrics[celltype + "_scprint"] = BenGRN(
            #    grn, doplot=False
            # ).scprint_benchmark()
            # del grn
            # gc.collect()
            grn_inferer = GNInfer(
                layer=layers,
                how="most var across",
                preprocess=(
                    "softpick"
                    if model.attention in ["softpick", "softpick-flash"]
                    else "softmax"
                ),
                head_agg="none",
                filtration="none",
                num_workers=8,
                num_genes=maxgenes,
                max_cells=maxcells,
                batch_size=batch_size,
            )
            grn = grn_inferer(model, adata[adata.X.sum(1) > 500], cell_type=celltype)
            grn.var.index = make_index_unique(grn.var["symbol"].astype(str))
            grn.varp["all"] = grn.varp["GRN"]
            grn.varp["GRN"] = grn.varp["GRN"].mean(-1)
            metrics[celltype + "_scprint_mean"] = BenGRN(
                grn, doplot=False
            ).scprint_benchmark()
            if clf_omni is None:
                grn.varp["GRN"] = grn.varp["all"]
                _, m, clf_omni = train_classifier(
                    grn,
                    C=1,
                    train_size=0.6,
                    max_iter=300,
                    class_weight={1: 800, 0: 1},
                    return_full=False,
                    shuffle=True,
                    doplot=False,
                )
                joblib.dump(clf_omni, "clf_omni.pkl")
                metrics["classifier"] = m
            coef = clf_omni.coef_[0] if clf_omni.coef_.shape[0] == 1 else clf_omni.coef_
            grn.varp["GRN"] = grn.varp["all"][:, :, coef > 0].mean(-1)
            metrics[celltype + "_scprint_class"] = BenGRN(
                grn, doplot=False
            ).scprint_benchmark()
            del grn
            gc.collect()
    return metrics


def softpick(x, dim=-1, eps=1e-5):
    # softpick function: relu(exp(x)-1) / sum(abs(exp(x)-1))
    # numerically stable version
    x_m = torch.max(x, dim=dim, keepdim=True).values
    x_m_e_m = torch.exp(-x_m)
    x_e_1 = torch.exp(x - x_m) - x_m_e_m
    r_x_e_1 = torch.nn.functional.relu(x_e_1)
    a_x_e_1 = torch.where(x.isfinite(), torch.abs(x_e_1), 0)
    return r_x_e_1 / (
        torch.sum(a_x_e_1, dim=dim, keepdim=True) + eps
    )  # epsilon is only useful if all inputs are EXACTLY 0. we might not even need it



"""
Finite Scalar Quantization: VQ-VAE Made Simple - https://arxiv.org/abs/2309.15505
Code adapted from Jax version in Appendix A.1
"""

from typing import List, Optional

import torch
import torch.nn as nn
from einops import pack, rearrange, unpack
from torch import Tensor, int32
from torch.nn import Module

# helper functions


def exists(v):
    return v is not None


def default(*args):
    for arg in args:
        if exists(arg):
            return arg
    return None


def pack_one(t, pattern):
    return pack([t], pattern)


def unpack_one(t, ps, pattern):
    return unpack(t, ps, pattern)[0]


# tensor helpers


def round_ste(z: Tensor) -> Tensor:
    """Round with straight through gradients."""
    zhat = z.round()
    return z + (zhat - z).detach()


# main class


class FSQ(Module):
    def __init__(
        self,
        levels: List[int],
        dim: Optional[int] = None,
        num_codebooks=1,
        keep_num_codebooks_dim: Optional[bool] = None,
        scale: Optional[float] = None,
    ):
        super().__init__()
        _levels = torch.tensor(levels, dtype=int32)
        self.register_buffer("_levels", _levels, persistent=False)

        _basis = torch.cumprod(torch.tensor([1] + levels[:-1]), dim=0, dtype=int32)
        self.register_buffer("_basis", _basis, persistent=False)

        self.scale = scale

        codebook_dim = len(levels)
        self.codebook_dim = codebook_dim

        effective_codebook_dim = codebook_dim * num_codebooks
        self.num_codebooks = num_codebooks
        self.effective_codebook_dim = effective_codebook_dim

        keep_num_codebooks_dim = default(keep_num_codebooks_dim, num_codebooks > 1)
        assert not (num_codebooks > 1 and not keep_num_codebooks_dim)
        self.keep_num_codebooks_dim = keep_num_codebooks_dim

        self.dim = default(dim, len(_levels) * num_codebooks)

        has_projections = self.dim != effective_codebook_dim
        self.project_in = (
            nn.Linear(self.dim, effective_codebook_dim)
            if has_projections
            else nn.Identity()
        )
        self.project_out = (
            nn.Linear(effective_codebook_dim, self.dim)
            if has_projections
            else nn.Identity()
        )
        self.has_projections = has_projections

        self.codebook_size = self._levels.prod().item()

        implicit_codebook = self.indices_to_codes(
            torch.arange(self.codebook_size), project_out=False
        )
        self.register_buffer("implicit_codebook", implicit_codebook, persistent=False)

    def bound(self, z: Tensor, eps: float = 1e-3) -> Tensor:
        """Bound `z`, an array of shape (..., d)."""
        half_l = (self._levels - 1) * (1 - eps) / 2
        offset = torch.where(self._levels % 2 == 0, 0.5, 0.0)
        shift = (offset / half_l).tan()
        return (z + shift).tanh() * half_l - offset

    def quantize(self, z: Tensor) -> Tensor:
        """Quantizes z, returns quantized zhat, same shape as z."""
        quantized = round_ste(self.bound(z))
        half_width = self._levels // 2  # Renormalize to [-1, 1].
        return quantized / half_width

    def _scale_and_shift(self, zhat_normalized: Tensor) -> Tensor:
        half_width = self._levels // 2
        return (zhat_normalized * half_width) + half_width

    def _scale_and_shift_inverse(self, zhat: Tensor) -> Tensor:
        half_width = self._levels // 2
        return (zhat - half_width) / half_width

    def codes_to_indices(self, zhat: Tensor) -> Tensor:
        """Converts a `code` to an index in the codebook."""
        assert zhat.shape[-1] == self.codebook_dim
        zhat = self._scale_and_shift(zhat)
        return (zhat * self._basis).sum(dim=-1).to(int32)

    def indices_to_codes(self, indices: Tensor, project_out=True) -> Tensor:
        """Inverse of `codes_to_indices`."""
        indices = rearrange(indices, "... -> ... 1")
        codes_non_centered = (indices // self._basis) % self._levels
        codes = self._scale_and_shift_inverse(codes_non_centered)

        if self.keep_num_codebooks_dim:
            codes = rearrange(codes, "... c d -> ... (c d)")

        if project_out:
            codes = self.project_out(codes)
        return codes

    def forward(self, z: Tensor) -> Tensor:
        """
        einstein notation
        b - batch
        n - sequence (or flattened spatial dimensions)
        d - feature dimension, which is also log2(codebook size)
        c - number of codebook dim
        """
        assert z.shape[-1] == self.dim, (
            f"expected dimension of {self.dim} but found dimension of {z.shape[-1]}"
        )
        small = self.project_in(z)

        z = rearrange(small, "b (c d) -> b c d", c=self.num_codebooks)

        codes = self.quantize(z)
        indices = self.codes_to_indices(codes)

        codes = rearrange(codes, "b c d -> b (c d)")

        out = self.project_out(codes)

        if not self.keep_num_codebooks_dim:
            indices = rearrange(indices, "... 1 -> ...")

        return out, indices, small


if __name__ == "__main__":
    levels = [8, 5, 5, 5]  # see 4.1 and A.4.1 in the paper
    quantizer = FSQ(levels)

    x = torch.randn(1, 4, 16, 16)  # 4 since there are 4 levels
    xhat, indices = quantizer(x)

    print(xhat.shape)  # (1, 1024, 4) - (batch, seq, dim)
    # print(indices.shape) # (1, 1024)    - (batch, seq)



"""Entry point for scprint."""

from lightning.pytorch import LightningModule, Trainer
from lightning.pytorch.cli import ArgsType, SaveConfigCallback
from lightning.pytorch.loggers import WandbLogger
from scdataloader import DataModule

from scprint2 import scPRINT2
from scprint2.cli import MyCLI


class MySaveConfig(SaveConfigCallback):
    """
    MySaveConfig is a subclass of SaveConfigCallback to parametrize the wandb logger further in cli mode
    """

    def setup(self, trainer: Trainer, pl_module: LightningModule, stage: str) -> None:
        if type(trainer.logger) is WandbLogger:
            if self.config.get("wandblog", "") != "":
                trainer.logger.watch(
                    pl_module,
                    log=self.config.get("wandblog", "all"),
                    log_freq=self.config.get("wandblog_freq", 1000),
                    log_graph=False,
                )
                if trainer.datamodule is not None:
                    trainer.logger.log_hyperparams({"datamodule": trainer.datamodule})
                    trainer.logger.log_hyperparams({"callbacks": trainer.callbacks})
            if trainer.is_global_zero:
                print(trainer.datamodule)
                print(trainer.callbacks)
                # print(pl_module.hparams)
        return super().setup(trainer, pl_module, stage)


def main(args: ArgsType = None):
    cli = MyCLI(  # noqa: F841
        scPRINT2,
        DataModule,
        args=args,
        parser_kwargs={"parser_mode": "omegaconf"},
        save_config_kwargs={"overwrite": True},
        save_config_callback=MySaveConfig,
    )


if __name__ == "__main__":  # pragma: no cover
    main()



import os

import pandas as pd
import torch
from torch import load

from scprint2.utils.utils import run_command

# https://github.com/agemagician/ProtTrans
# https://academic.oup.com/nargab/article/4/1/lqac012/6534363


class ESM2:
    def __init__(
        self,
        config: str = "esm-extract",
        pretrained_model: str = "esm2_t33_650M_UR50D",
    ):
        """
        ESM2 a ghost class to call protein LLMs to encode protein sequences.

        Args:
            config (str, optional): The configuration for the model. Defaults to "esm-extract".
            pretrained_model (str, optional): The pretrained model to be used. Defaults to "esm2_t33_650M_UR50D".
        """
        self.config = config
        self.pretrained_model = pretrained_model

    def __call__(
        self, input_file: str, output_folder: str = "/tmp/esm_out/", cache: bool = True
    ) -> pd.DataFrame:
        """
        Call the ESM2 model on the input file.

        Args:
            input_file (str): The input file to be processed.
            output_folder (str, optional): The folder where the output will be stored. Defaults to "/tmp/esm_out/".
            cache (bool, optional): If True, use cached data if available. Defaults to True.

        Returns:
            pd.DataFrame: The results of the model as a DataFrame.
        """

        if not os.path.exists(output_folder) or not cache:
            os.makedirs(output_folder, exist_ok=True)
            print("running ESM2")
            cmd = (
                self.config
                + " "
                + self.pretrained_model
                + " "
                + input_file
                + " "
                + output_folder
                + " --include mean"
            )
            try:
                run_command(cmd, shell=True)
            except Exception as e:
                raise RuntimeError(
                    "An error occurred while running the esm-extract command: " + str(e)
                )
        return self.read_results(output_folder)

    def read_results(self, output_folder: str) -> pd.DataFrame:
        """
        Read multiple .pt files in a folder and convert them into a DataFrame.

        Args:
            output_folder (str): The folder where the .pt files are stored.

        Returns:
            pd.DataFrame: The results of the model as a DataFrame.
        """
        files = os.listdir(output_folder)
        files = [i for i in files if i.endswith(".pt")]
        results = []
        for file in files:
            results.append(
                load(output_folder + file)["mean_representations"][33].numpy().tolist()
            )
        return pd.DataFrame(data=results, index=[file.split(".")[0] for file in files])
